#!/usr/bin/env python

import logging
import re
import subprocess
from pathlib import Path
from typing import List, Optional, Union

from ltpylib import files, procs

SHEBANG_REGEX = re.compile(r"^#!(.*)$")
FILE_EXT_MAPPINGS = {
  "jsonl": "json",
  "json5": "json",
  "lookml": "yaml",
  "py": "python",
  "rb": "ruby",
  "sh": "bash",
  "yml": "yaml",
}
FILE_NAME_MAPPINGS = {
  "Gemfile": "ruby",
}


def prettify(
  files_to_prettify: Union[Path, List[Path]],
  file_type: str = None,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
  read_shebang_if_necessary: bool = True,
):
  if not isinstance(files_to_prettify, list):
    files_to_prettify = [files_to_prettify]

  for file in files_to_prettify:
    if file_type:
      single_file_type = file_type
    else:
      single_file_type = FILE_EXT_MAPPINGS.get(file.suffix[1:], file.suffix[1:])
      if not single_file_type and read_shebang_if_necessary:
        single_file_type = read_file_shebang(file)

      if not single_file_type:
        single_file_type = FILE_NAME_MAPPINGS.get(file.name, None)

    if not single_file_type:
      raise ValueError("Could not determine file type: file=%s" % file.as_posix())

    func_for_type = globals()["prettify_" + single_file_type + "_file"]
    if not callable(func_for_type):
      raise ValueError("Unsupported file type: file=%s type=%s" % (file.as_posix(), single_file_type))

    func_for_type(file, compact=compact, debug_mode=debug_mode, verbose=verbose)
    if verbose:
      logging.debug("Updated %s", file.as_posix())


def prettify_bash_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  formatter_args = [
    "shfmt",
    "--simplify",
    "--indent",
    "2",
    "--case-indent",
    "--write",
  ]

  if compact:
    formatter_args.append("--minify")

  formatter_args.append(file.as_posix())

  run_formatter(
    file,
    formatter_args,
    debug_mode=debug_mode,
    verbose=verbose,
    use_run_with_regular_stdout=True,
    should_have_stdout=False,
  )


def prettify_html_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  formatter_path = "/usr/bin/tidy" if Path("/usr/bin/tidy").exists() else "tidy"
  formatter_args = [
    formatter_path,
    "-icm",
    "-wrap",
    "200",
    "--doctype",
    "omit",
    "--indent",
    "yes",
    "--vertical-space",
    "no",
  ]

  if not verbose:
    formatter_args.extend([
      "-quiet",
      "--show-warnings",
      "no",
    ])

  formatter_args.append(file.as_posix())

  run_formatter(
    file,
    formatter_args,
    debug_mode=debug_mode,
    verbose=verbose,
    use_run_with_regular_stdout=True,
    should_have_stdout=False,
    allow_exit_codes=[0, 1],
  )


def prettify_json_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  jq_args = ["--sort-keys", ".", file.as_posix()]
  if compact:
    jq_args.insert(0, "--compact-output")

  result = run_formatter(
    file,
    ["jq"] + jq_args,
    debug_mode=debug_mode,
    verbose=verbose,
  )
  files.write_file(file, result.stdout)


def prettify_python_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  formatter_args = [
    "yapf",
    "--in-place",
    "--recursive",
  ]
  yapf_style_home_file = Path.home().joinpath(".style.yapf")
  if yapf_style_home_file.is_file():
    formatter_args.extend([
      "--style",
      yapf_style_home_file.as_posix(),
    ])

  formatter_args.append(file.as_posix())
  run_formatter(
    file,
    formatter_args,
    debug_mode=debug_mode,
    verbose=verbose,
    should_have_stdout=False,
    use_run_with_regular_stdout=True,
  )


def prettify_ruby_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  formatter_args = [
    "standardrb",
    "--fix",
    "--",
    file.as_posix(),
  ]

  run_formatter(
    file,
    formatter_args,
    debug_mode=debug_mode,
    verbose=verbose,
    should_have_stdout=False,
    use_run_with_regular_stdout=True,
  )


def prettify_sql_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  formatter_args = [
    "sql-formatter",
    "--language",
    "sqlite",
    "--config",
    Path.home().joinpath(".config/sql-formatter/sqlite.json").as_posix(),
    "--output",
    file.as_posix(),
    file.as_posix(),
  ]
  run_formatter(
    file,
    formatter_args,
    debug_mode=debug_mode,
    verbose=verbose,
    use_run_with_regular_stdout=True,
  )


def prettify_xml_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  result = run_formatter(
    file,
    ["xmllint", "--format", file.as_posix()],
    debug_mode=debug_mode,
    verbose=verbose,
  )
  files.write_file(file, result.stdout)


def prettify_yaml_file(
  file: Path,
  compact: bool = False,
  debug_mode: bool = False,
  verbose: bool = False,
):
  result = run_formatter(
    file,
    ["yq", "--yaml-roundtrip", "--indentless-lists", "--sort-keys", "--width=5000", ".", file.as_posix()],
    debug_mode=debug_mode,
    verbose=verbose,
  )
  files.write_file(file, result.stdout)


def run_formatter(
  file: Path,
  formatter_args: List[str],
  debug_mode: bool = False,
  verbose: bool = False,
  use_run_with_regular_stdout: bool = False,
  should_have_stdout: bool = True,
  allow_exit_codes: List[int] = None,
) -> subprocess.CompletedProcess:
  result = check_proc_result(
    file,
    procs.run_with_regular_stdout(formatter_args, log_cmd=verbose) if use_run_with_regular_stdout else procs.run(formatter_args, log_cmd=verbose),
    should_have_stdout=should_have_stdout,
    allow_exit_codes=allow_exit_codes,
  )
  return result


def check_proc_result(
  file: Path,
  result: subprocess.CompletedProcess,
  should_have_stdout: bool = True,
  allow_exit_codes: List[int] = None,
) -> subprocess.CompletedProcess:
  exit_code = result.returncode
  proc_succeeded = exit_code == 0
  if allow_exit_codes and exit_code in allow_exit_codes:
    proc_succeeded = True

  if should_have_stdout and not result.stdout:
    raise Exception("Issue prettifying file: file=%s status=%s stderr=%s" % (file.as_posix(), exit_code, result.stderr))

  if result.stderr:
    if proc_succeeded:
      logging.warning(result.stderr)
    else:
      raise Exception("Issue prettifying file: file=%s status=%s stderr=%s stdout=%s" % (file.as_posix(), exit_code, result.stderr, result.stderr))

  if not proc_succeeded:
    result.check_returncode()

  return result


def read_file_shebang(file: Path) -> Optional[str]:
  if file.is_file():
    file_lines = files.read_file_n_lines(file, n_lines=1)
    if file_lines and len(file_lines) >= 1:
      first_line = file_lines[0]
      match = SHEBANG_REGEX.fullmatch(first_line)
      if match:
        full_command = match.group(1)
        command_parts = full_command.split(" ")

        shebang_file_type = ""
        if len(command_parts) == 1 or not command_parts[0].endswith("env"):
          shebang_file_type = Path(command_parts[0]).name
        else:
          shebang_file_type = command_parts[1]

        logging.debug("Resolved file type from shebang: file=%s type=%s first_line=%s", file.as_posix(), shebang_file_type, first_line)
        return shebang_file_type
      else:
        logging.debug("Shebang regex did not match: file=%s regex=%s first_line=%s", file.as_posix(), SHEBANG_REGEX, first_line)
    else:
      logging.debug("File is empty: file=%s", file.as_posix())



#!/usr/bin/env python
import logging
import re
import subprocess
import sys
from pathlib import Path
from types import TracebackType
from typing import Any, Callable, IO, Iterable, List, Optional, Tuple, Type, Union

CMD_NO_QUOTES_NEEDED_REGEX = re.compile(r"^[a-zA-Z0-9_-]+$")


class CalledProcessErrorWithOutput(subprocess.CalledProcessError):

  def __init__(self, returncode, cmd, output=None, stderr=None):
    super(CalledProcessErrorWithOutput, self).__init__(returncode, cmd, output, stderr)

  def __str__(self):
    stdout_and_stderr = ""

    if self.output:
      stdout_and_stderr += "\n\nstdout\n---\n%s" % self.output

    if self.stderr:
      stdout_and_stderr += "\n\nstderr\n---\n%s" % self.stderr

    if self.returncode and self.returncode < 0:
      try:
        import signal

        return "Command '%s' died with %r.%s" % (self.cmd, signal.Signals(-self.returncode), stdout_and_stderr)
      except ValueError:
        return "Command '%s' died with unknown signal %d.%s" % (self.cmd, -self.returncode, stdout_and_stderr)
    else:
      return "Command '%s' returned non-zero exit status %d.%s" % (self.cmd, self.returncode, stdout_and_stderr)


def run_popen(
  *popenargs,
  cwd: Union[str, bytes, Path] = None,
  shell: bool = False,
  stdout: Optional[Union[int, IO]] = subprocess.PIPE,
  stderr: Optional[Union[int, IO]] = subprocess.PIPE,
  **kwargs,
) -> subprocess.Popen:
  kwargs['universal_newlines'] = True

  return subprocess.Popen(
    *popenargs,
    cwd=cwd,
    shell=shell,
    stdout=stdout,
    stderr=stderr,
    **kwargs,
  )


def run(
  *popenargs,
  input: Union[bytes, str, None] = None,
  timeout: Optional[float] = None,
  check: bool = False,
  cwd: Union[str, bytes, Path] = None,
  shell: bool = False,
  stdout: Optional[Union[int, IO]] = subprocess.PIPE,
  stderr: Optional[Union[int, IO]] = subprocess.PIPE,
  log_cmd: bool = False,
  log_cmd_level: int = logging.INFO,
  **kwargs,
) -> subprocess.CompletedProcess:
  if log_cmd:
    from ltpylib import logs

    logs.log_title_with_sep(
      create_cmd_debug_string(popenargs[0]) if isinstance(popenargs[0], Iterable) else popenargs[0],
      level=log_cmd_level,
    )

  kwargs['universal_newlines'] = True

  result = subprocess.run(
    *popenargs,
    input=input,
    timeout=timeout,
    check=False,
    cwd=cwd,
    shell=shell,
    stdout=stdout,
    stderr=stderr,
    **kwargs,
  )

  if check:
    check_returncode_with_output(result)

  return result


def run_with_logging_output(
  *popenargs,
  input: Union[bytes, str, None] = None,
  timeout: Optional[float] = None,
  check: bool = False,
  cwd: Union[str, bytes, Path] = None,
  shell: bool = False,
  level: int = logging.INFO,
  log_cmd: bool = False,
  log_cmd_level: int = logging.INFO,
  **kwargs,
) -> subprocess.CompletedProcess:
  from ltpylib import logs

  with logs.LogPipe(level=level) as log_pipe:
    return run(
      *popenargs,
      input=input,
      timeout=timeout,
      check=check,
      cwd=cwd,
      log_cmd=log_cmd,
      log_cmd_level=log_cmd_level,
      shell=shell,
      stdout=log_pipe,
      stderr=log_pipe,
      **kwargs,
    )


def run_with_regular_stdout(
  *popenargs,
  input: Union[bytes, str, None] = None,
  timeout: Optional[float] = None,
  check: bool = False,
  cwd: Union[str, bytes, Path] = None,
  shell: bool = False,
  log_cmd: bool = False,
  log_cmd_level: int = logging.INFO,
  **kwargs,
) -> subprocess.CompletedProcess:
  return run(
    *popenargs,
    input=input,
    timeout=timeout,
    check=check,
    cwd=cwd,
    log_cmd=log_cmd,
    log_cmd_level=log_cmd_level,
    shell=shell,
    stdout=sys.stdout,
    stderr=sys.stderr,
    **kwargs,
  )


def run_and_parse_output(
  *popenargs,
  input: Union[bytes, str, None] = None,
  timeout: Optional[float] = None,
  check: bool = False,
  cwd: Union[str, bytes, Path] = None,
  shell: bool = False,
  log_cmd: bool = False,
  log_cmd_level: int = logging.INFO,
  **kwargs,
) -> Tuple[int, str]:
  kwargs['stdout'] = subprocess.PIPE

  result = run(
    *popenargs,
    input=input,
    timeout=timeout,
    check=check,
    cwd=cwd,
    log_cmd=log_cmd,
    log_cmd_level=log_cmd_level,
    shell=shell,
    **kwargs,
  )

  return result.returncode, result.stdout


def run_and_parse_output_on_success(
  *popenargs,
  input: Union[bytes, str, None] = None,
  timeout: Optional[float] = None,
  check: bool = True,
  cwd: Union[str, bytes, Path] = None,
  shell: bool = False,
  log_cmd: bool = False,
  log_cmd_level: int = logging.INFO,
  **kwargs,
) -> str:
  kwargs['stdout'] = subprocess.PIPE

  result = run(
    *popenargs,
    input=input,
    timeout=timeout,
    check=check,
    cwd=cwd,
    log_cmd=log_cmd,
    log_cmd_level=log_cmd_level,
    shell=shell,
    **kwargs,
  )

  return result.stdout


def check_returncode_with_output(result: subprocess.CompletedProcess):
  if result.returncode:
    raise CalledProcessErrorWithOutput(result.returncode, result.args, result.stdout, result.stderr)


def get_procs_from_name(name_matcher: str) -> List[Tuple[int, str]]:
  matched_procs = []
  exit_code, output = run_and_parse_output(['pgrep', '-fl', name_matcher])
  if exit_code > 0:
    return matched_procs

  for line in output.splitlines():
    parts = line.partition(' ')
    matched_procs.append((int(parts[0]), parts[2]))

  return matched_procs


def proc_debug_string(proc) -> str:
  """
  :type proc: psutil.Process
  """
  import collections
  import psutil

  info = collections.OrderedDict()
  info['pid'] = proc.pid
  try:
    info["name"] = proc.name()
    if proc._create_time:
      info['started'] = psutil._pprint_secs(proc._create_time)
  except psutil.ZombieProcess:
    info["status"] = "zombie"
  except psutil.NoSuchProcess:
    info["status"] = "terminated"
  except psutil.AccessDenied:
    pass
  info['cmdline'] = ' '.join(proc.cmdline())
  return "%s.%s(%s)" % (
    proc.__class__.__module__,
    proc.__class__.__name__,
    ", ".join(["%s=%r" % (k, v) for k, v in info.items()]),
  )


def create_cmd_debug_string(popenargs: Iterable[str]) -> str:
  output: List[str] = []
  for part in popenargs:
    if CMD_NO_QUOTES_NEEDED_REGEX.fullmatch(part):
      output.append(part)
    elif "'" in part:
      escaped_part = part.replace('"', '\\"')
      output.append(f"\"{escaped_part}\"")
    else:
      output.append(f"'{part}'")

  return " ".join(output)


def stop_proc_by_name(name_matcher: str) -> bool:
  import psutil

  matched_procs = get_procs_from_name(name_matcher)
  if not matched_procs:
    return False

  if len(matched_procs) > 1:
    raise Exception('Multiple processes found: %s' % matched_procs)

  proc = psutil.Process(matched_procs[0][0])
  proc.terminate()
  return True


def stop_proc_by_pid(pid: Any) -> bool:
  import psutil

  proc = psutil.Process(int(pid))
  if proc.is_running():
    proc.terminate()
  else:
    return False
  return True


def await_termination(pid: int, timeout: int = 30, sleep_time: int = 1, log_level: int = 20):
  """
  :param log_level: default is logging.INFO
  :return: None
  """
  import logging
  import psutil
  from time import sleep

  if pid is None:
    return

  proc = psutil.Process(pid)
  total_time = 0
  while total_time <= timeout:
    if proc.is_running():
      logging.log(log_level, 'STATUS: loop %s seconds - process is still running, please wait...', total_time)
    else:
      break

    total_time += sleep_time
    sleep(sleep_time)

  if proc.is_running():
    logging.error('STATUS: Process still running after %s seconds, sending SIGKILL to terminate...', total_time)
    proc.kill()
    proc.wait(5)
    logging.log(log_level, 'STATUS: Process killed.')
  else:
    logging.log(log_level, 'STATUS: Process successfully shutdown after %s seconds.', total_time)


orig_unraisablehook = None


def disable_unraisablehook():
  if not hasattr(sys, "unraisablehook"):
    return

  global orig_unraisablehook
  if orig_unraisablehook is None:
    orig_unraisablehook = sys.unraisablehook

  sys.unraisablehook = ignoring_unraisablehook


def enable_unraisablehook():
  if not hasattr(sys, "unraisablehook"):
    return

  global orig_unraisablehook
  if orig_unraisablehook is None:
    raise Exception("sys.unraisablehook was never overridden")
  sys.unraisablehook = orig_unraisablehook
  orig_unraisablehook = None


def ignoring_unraisablehook(*args, **kwargs):
  pass


class catch_unraisable_exception:

  def __init__(self) -> None:
    self.unraisable: Optional["sys.UnraisableHookArgs"] = None
    self._old_hook: Optional[Callable[["sys.UnraisableHookArgs"], Any]] = None

  def _hook(self, unraisable: "sys.UnraisableHookArgs") -> None:
    # Storing unraisable.object can resurrect an object which is being
    # finalized. Storing unraisable.exc_value creates a reference cycle.
    self.unraisable = unraisable

  def __enter__(self) -> "catch_unraisable_exception":
    self._old_hook = sys.unraisablehook
    sys.unraisablehook = self._hook
    return self

  def __exit__(
    self,
    exc_type: Optional[Type[BaseException]],
    exc_val: Optional[BaseException],
    exc_tb: Optional[TracebackType],
  ) -> None:
    assert self._old_hook is not None
    sys.unraisablehook = self._old_hook
    self._old_hook = None
    del self.unraisable


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
import os
import unittest
from pathlib import Path

from ltpylib import files


class TestFiles(unittest.TestCase):

  def test_read_file_n_lines(self):
    test_file = Path(os.path.dirname(os.path.realpath(__file__))).joinpath("resources/test_files_read_file_n_lines.txt")

    assert files.read_file_n_lines(test_file, 1) == ["line1"]
    assert files.read_file_n_lines(test_file, 2) == ["line1", "line2"]
    assert files.read_file_n_lines(test_file, 3) == ["line1", "line2", "line3"]
    assert files.read_file_n_lines(test_file, 4) == ["line1", "line2", "line3", ""]
    assert files.read_file_n_lines(test_file, 5) == ["line1", "line2", "line3", "", "line5"]
    assert files.read_file_n_lines(test_file, 6) == ["line1", "line2", "line3", "", "line5"]
    assert files.read_file_n_lines(test_file, -1) == ["line1", "line2", "line3", "", "line5"]

  def test_read_json_file(self):
    test_file = Path(os.path.dirname(os.path.realpath(__file__))).joinpath("resources/test_files_read_json_file.json")
    parsed_json = files.read_json_file(test_file)

    assert parsed_json.get("field1") == "value1"
    assert parsed_json.get("field2") == 2


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# flake8: noqa E722
import os
import platform
import shlex
import struct
import subprocess
from typing import Tuple


def get_terminal_size() -> Tuple[int, int]:
  """ getTerminalSize()
   - get width and height of console
   - works on linux,os x,windows,cygwin(windows)
   originally retrieved from:
   http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
  """
  current_os = platform.system()
  tuple_xy = None
  if current_os == 'Windows':
    tuple_xy = _get_terminal_size_windows()
    if tuple_xy is None:
      tuple_xy = _get_terminal_size_tput()
      # needed for window's python in cygwin's xterm!
  if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
    tuple_xy = _get_terminal_size_linux()
  if tuple_xy is None:
    tuple_xy = (200, 25)
  return tuple_xy


def _get_terminal_size_windows():
  try:
    from ctypes import windll, create_string_buffer

    # stdin handle is -10
    # stdout handle is -11
    # stderr handle is -12
    h = windll.kernel32.GetStdHandle(-12)
    csbi = create_string_buffer(22)
    res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
    if res:
      (bufx, bufy, curx, cury, wattr, left, top, right, bottom, maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
      sizex = right - left + 1
      sizey = bottom - top + 1
      return sizex, sizey
  except:
    pass


def _get_terminal_size_tput():
  try:
    cols = int(subprocess.check_call(shlex.split('tput cols')))
    rows = int(subprocess.check_call(shlex.split('tput lines')))
    return (cols, rows)
  except:
    pass


def _get_terminal_size_linux():

  def ioctl_gwinsz(fd):
    try:
      import fcntl
      import termios

      maybe_cols_rows = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
      return maybe_cols_rows
    except:
      pass

  cols_rows = ioctl_gwinsz(0) or ioctl_gwinsz(1) or ioctl_gwinsz(2)
  if not cols_rows:
    try:
      file_desc = os.open(os.ctermid(), os.O_RDONLY)
      cols_rows = ioctl_gwinsz(file_desc)
      os.close(file_desc)
    except:
      pass
  if not cols_rows:
    try:
      cols_rows = (os.environ['LINES'], os.environ['COLUMNS'])
    except:
      return None
  return int(cols_rows[1]), int(cols_rows[0])


if __name__ == "__main__":
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)



#!/usr/bin/env python
import contextlib
import logging
import re
from dataclasses import dataclass
from typing import List, Optional

from ltpylib import strings
from ltpylib.common_types import TypeWithDictRepr


class ChromeExtensions(TypeWithDictRepr):

  def __init__(self):
    self.installed: List[ChromeExtension] = []
    self.pinned_extensions: List[str] = []
    self.toolbar: List[str] = []


@dataclass
class ChromeExtension:
  extension_id: str
  name: str
  description: str
  stars: Optional[float] = None
  ratings: Optional[int] = None


def create_chrome_extension(extension_id: str, use_search: bool = False) -> ChromeExtension:
  import requests
  from bs4 import BeautifulSoup
  from ltpylib import htmlparser

  if use_search:
    url = f"https://chromewebstore.google.com/search/{extension_id}?hl=en-US"
  else:
    url = f"https://chromewebstore.google.com/detail/{extension_id}"

  logging.debug(f"Getting extension details: {url}")
  response = requests.get(url, headers={"User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36"})
  parsed_html = htmlparser.create_parser(response.text)

  if use_search:
    name_elem = parsed_html.select_one(f"[data-item-id=\"{extension_id}\"] h2")
    description_elem = parsed_html.select_one(f"[data-item-id=\"{extension_id}\"] > div > div > p:last-child")
    name = name_elem.text if name_elem else None
    description = description_elem.text if description_elem else None
    stars = None
    ratings = None
  else:
    name_elem = parsed_html.select_one("meta[property='og:title']")
    description_elem = parsed_html.select_one("meta[property='og:description']")
    ratings_elem = parsed_html.select_one("a[href*='/reviews'] > p")
    stars_elem: BeautifulSoup | None = None
    if ratings_elem:
      try:
        stars_elem = next(ratings_elem.parent.parent.parent.children)
      except:
        pass

    name = name_elem.get("content") if name_elem else None
    description = description_elem.get("content") if description_elem else None
    with contextlib.suppress(BaseException):
      stars = strings.convert_to_number(stars_elem.text) if stars_elem else None
    with contextlib.suppress(BaseException):
      ratings = strings.convert_to_number(re.search(r"\d+", ratings_elem.text).group()) if ratings_elem else None

  return ChromeExtension(
    extension_id=extension_id,
    name=name,
    description=description,
    stars=stars,
    ratings=ratings,
  )



#!/usr/bin/env python


class TermColors:
  BLACK = '\033[0;30m'
  BLUE = '\033[0;34m'
  BOLD = '\033[1m'
  CYAN = '\033[0;36m'
  ENDC = '\033[0m'
  GREEN = '\033[0;32m'
  HEADER = '\033[95m'
  LIGHT_BLUE = '\033[1;34m'
  LIGHT_CYAN = '\033[1;36m'
  LIGHT_GREEN = '\033[1;32m'
  LIGHT_PURPLE = '\033[1;35m'
  LIGHT_RED = '\033[1;31m'
  LIGHT_YELLOW = '\033[1;33m'
  PURPLE = '\033[0;35m'
  RED = '\033[0;31m'
  UNDERLINE = '\033[4m'
  WHITE = '\033[0;37m'
  YELLOW = '\033[0;33m'


def black(value: str) -> str:
  return TermColors.BLACK + value + TermColors.ENDC


def blue(value: str) -> str:
  return TermColors.BLUE + value + TermColors.ENDC


def bold(value: str) -> str:
  return TermColors.BOLD + value + TermColors.ENDC


def cyan(value: str) -> str:
  return TermColors.CYAN + value + TermColors.ENDC


def green(value: str) -> str:
  return TermColors.GREEN + value + TermColors.ENDC


def light_blue(value: str) -> str:
  return TermColors.LIGHT_BLUE + value + TermColors.ENDC


def light_cyan(value: str) -> str:
  return TermColors.LIGHT_CYAN + value + TermColors.ENDC


def light_green(value: str) -> str:
  return TermColors.LIGHT_GREEN + value + TermColors.ENDC


def light_purple(value: str) -> str:
  return TermColors.LIGHT_PURPLE + value + TermColors.ENDC


def light_red(value: str) -> str:
  return TermColors.LIGHT_RED + value + TermColors.ENDC


def light_yellow(value: str) -> str:
  return TermColors.LIGHT_YELLOW + value + TermColors.ENDC


def purple(value: str) -> str:
  return TermColors.PURPLE + value + TermColors.ENDC


def red(value: str) -> str:
  return TermColors.RED + value + TermColors.ENDC


def underline(value: str) -> str:
  return TermColors.UNDERLINE + value + TermColors.ENDC


def white(value: str) -> str:
  return TermColors.WHITE + value + TermColors.ENDC


def yellow(value: str) -> str:
  return TermColors.YELLOW + value + TermColors.ENDC



#!/usr/bin/env python
import os
import tempfile
from pathlib import Path
from typing import Tuple, Union

from ltpylib import files, procs
from ltpylib.files import convert_to_path


def diff_git(
  initial: str,
  updated: str,
  add_suffix: str = None,
  context_lines: int = None,
  color: str = "always",
) -> Tuple[str, bool]:
  suffix_prefix = ("." + add_suffix) if add_suffix else ""

  ifd, initial_temp_file = tempfile.mkstemp(suffix=suffix_prefix + ".initial")
  ufd, updated_temp_file = tempfile.mkstemp(suffix=suffix_prefix + ".updated")
  os.close(ifd)
  os.close(ufd)

  files.write_file(initial_temp_file, initial)
  files.write_file(updated_temp_file, updated)

  try:
    return diff_git_files(initial_temp_file, updated_temp_file, context_lines=context_lines, color=color)
  finally:
    os.remove(initial_temp_file)
    os.remove(updated_temp_file)


def diff_git_files(
  initial_file: Union[Path, str],
  updated_file: Union[Path, str],
  context_lines: int = None,
  color: str = "always",
) -> Tuple[str, bool]:
  initial_file = convert_to_path(initial_file)
  updated_file = convert_to_path(updated_file)

  command = [
    "git",
    "diff",
    "--no-index",
    "-w",
  ]

  if color:
    command.append("--color=%s" % color)

  if context_lines is not None:
    command.append("--unified=%s" % str(context_lines))

  command.extend([initial_file, updated_file])
  result = procs.run(
    command,
    check=False,
  )
  return result.stdout, result.returncode == 1



#!/usr/bin/env python
from typing import List

import semver

STABLE_PRERELEASE_VERSIONS = ["final", "release"]


class LenientVersionInfo(semver.VersionInfo):

  def __init__(self, major: int, minor: int = 0, patch: int = 0, prerelease: str = None, build: str = None, original_version: str = None):
    self.original_version: str = original_version
    semver.VersionInfo.__init__(self, major, minor, patch, prerelease, build)
    self.version: str = self.__str__()

  def is_stable(self) -> bool:
    return self.build is None and (self.prerelease is None or self.prerelease.lower() in STABLE_PRERELEASE_VERSIONS)

  @staticmethod
  def make_version_semantic(version: str) -> str:
    dot_count = version.count(".")
    if dot_count <= 0:
      version = version + ".0.0"
    elif dot_count <= 1:
      version = version + ".0"

    return version

  @staticmethod
  def parse_lenient(original_version: str) -> 'LenientVersionInfo':
    version = LenientVersionInfo.make_version_semantic(original_version)

    prerelease = None
    try:
      parsed = semver.VersionInfo.parse(version)
      prerelease = parsed.prerelease
    except ValueError:
      parts = original_version.split(".")
      if len(parts) > 3:
        prerelease = ".".join(parts[3:])
      elif "-" in original_version:
        main_part, prerelease = original_version.split("-", maxsplit=1)
        parts = LenientVersionInfo.make_version_semantic(main_part).split(".")
      else:
        raise

      parsed = semver.VersionInfo.parse(".".join(parts[:3]))

    return LenientVersionInfo(
      parsed.major,
      minor=parsed.minor,
      patch=parsed.patch,
      prerelease=prerelease,
      build=parsed.build,
      original_version=original_version,
    )


def parse_semver_lenient(version: str) -> LenientVersionInfo:
  return LenientVersionInfo.parse_lenient(version)


def parse_all_semver_lenient(
  versions: List[str],
  ignore_invalid: bool = False,
  remove_non_stable: bool = False,
) -> List[LenientVersionInfo]:
  parsed_versions: List[LenientVersionInfo] = []
  for v in versions:
    try:
      parsed = LenientVersionInfo.parse_lenient(v)
      if remove_non_stable and not parsed.is_stable():
        continue

      parsed_versions.append(parsed)
    except:  # noqa: E722
      if ignore_invalid:
        continue

      raise

  return parsed_versions



#!/usr/bin/env python
# pylint: disable=C0111
from typing import Any, Sequence


def check_command(cmd: str) -> bool:
  import shutil

  return shutil.which(cmd) is not None


def is_empty(val) -> bool:
  if val is None:
    return True

  if isinstance(val, (dict, list)):
    return not val

  return False


def is_not_empty(val) -> bool:
  return not is_empty(val)


def is_any_value_present(check: Sequence[Any], values: Sequence[Any]) -> bool:
  if not check or not values:
    return False

  for val in values:
    if val in check:
      return True

  return False



#!/usr/bin/env python
import logging
import time
from typing import Callable, SupportsFloat


def format_seconds(frac_seconds):
  hours, rem = divmod(float(frac_seconds), 3600)
  minutes, seconds = divmod(rem, 60)
  return "{0:0>2}:{1:0>2}:{2:06.3f}".format(int(hours), int(minutes), seconds)


def format_millis(millis):
  return format_seconds(float(millis) / float(1000))


def get_time_elapsed_msg(start_time):
  return format_seconds(time.time() - float(start_time))


def get_time_remaining_formatted_seconds(start_time, count, total):
  frac_seconds = time.time() - float(start_time)
  if float(count) <= 0:
    return "N/A"

  estimated_total = frac_seconds * (float(total) / float(count))
  return format_seconds(estimated_total - frac_seconds)


def get_time_remaining_msg(start_time, count, total):
  frac_seconds = time.time() - float(start_time)
  if float(count) <= 0:
    return "Elapsed: {0: >12} Remaining: {1: >12}".format(format_seconds(frac_seconds), 'N/A')

  estimated_total = frac_seconds * (float(total) / float(count))
  return "Elapsed: {0: >12} Remaining: {1: >12}".format(format_seconds(frac_seconds), format_seconds(estimated_total - frac_seconds))


def sleep_and_log(seconds: int, log_level: int = logging.INFO):
  logging.log(log_level, "Sleeping %s seconds...", seconds)
  time.sleep(seconds)


def has_time_remaining(
  start_time: float,
  max_seconds: SupportsFloat,
  log_time_ran_out: bool = False,
  time_ran_out_func: Callable[[float, SupportsFloat, float], None] = None,
) -> bool:
  curr_time = time.time()
  check_result = (curr_time - start_time) < float(max_seconds)

  if not check_result:
    if log_time_ran_out:
      logging.warning(f"Timed out waiting after {max_seconds:,} seconds")

    if time_ran_out_func is not None:
      time_ran_out_func(start_time, max_seconds, curr_time)

  return check_result


if __name__ == "__main__":
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)



#!/usr/bin/env python
from datetime import datetime
from enum import auto
from typing import Dict, List, Optional

from ltpylib import dates, enums
from ltpylib.common_types import DataWithUnknownPropertiesAsAttributes


def in_values_and_valid_date(values: dict, key: str) -> bool:
  return key in values and values.get(key) != "None"


class IdAndSelf(object):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.id: int = int(values.pop("id")) if "id" in values else None
    self.self: str = values.pop("self", None)


class NameIdAndSelf(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.name: str = values.pop("name", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class ValueIdAndSelf(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.value: str = values.pop("value", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class TextAndValue(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.text: str = values.pop("text", None)
    self.value: float = float(values.pop("value")) if "value" in values else None

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class FixVersion(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.archived: bool = values.pop("archived", None)
    self.name: str = values.pop("name", None)
    self.released: bool = values.pop("released", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class Issue(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.acceptanceCriteria: str = values.pop("acceptanceCriteria", None)
    self.aggregateprogress: Dict[str, int] = values.pop("aggregateprogress", None)
    self.aggregatetimeestimate: int = values.pop("aggregatetimeestimate", None)
    self.aggregatetimeoriginalestimate: int = values.pop("aggregatetimeoriginalestimate", None)
    self.aggregatetimespent: int = values.pop("aggregatetimespent", None)
    self.assignee: JiraUser = JiraUser(values=values.pop("assignee")) if "assignee" in values else None
    self.attachment: List[dict] = values.pop("attachment", None)
    self.comment: Dict[str, int] = values.pop("comment", None)
    self.components: List[NameIdAndSelf] = list(map(NameIdAndSelf, values.pop("components", []))) if "components" in values else None
    self.controlGroup: ValueIdAndSelf = ValueIdAndSelf(values=values.pop("controlGroup")) if "controlGroup" in values else None
    self.created: datetime = dates.parse_iso_date(values.pop("created")) if "created" in values else None
    self.creator: JiraUser = JiraUser(values=values.pop("creator")) if "creator" in values else None
    self.description: str = values.pop("description", None)
    self.development: str = values.pop("development", None)
    self.epicColour: str = values.pop("epicColour", None)
    self.epicLink: str = values.pop("epicLink", None)
    self.epicName: str = values.pop("epicName", None)
    self.epicStatus: ValueIdAndSelf = ValueIdAndSelf(values=values.pop("epicStatus")) if "epicStatus" in values else None
    self.expand: str = values.pop("expand", None)
    self.fixVersions: List[FixVersion] = list(map(FixVersion, values.pop("fixVersions", []))) if "fixVersions" in values else None
    self.issuelinks: List[IssueLink] = list(map(IssueLink, values.pop("issuelinks", []))) if "issuelinks" in values else None
    self.issuetype: IssueType = IssueType(values=values.pop("issuetype")) if "issuetype" in values else None
    self.key: str = values.pop("key", None)
    self.labels: List[str] = values.pop("labels", None)
    self.lastViewed: datetime = dates.parse_date(values.pop("lastViewed")) if "lastViewed" in values else None
    self.names: Dict[str, str] = values.pop("names", None)
    self.onsite: ValueIdAndSelf = ValueIdAndSelf(values=values.pop("onsite")) if "onsite" in values else None
    self.parent: IssueBasic = IssueBasic(values=values.pop("parent")) if "parent" in values else None
    self.planned: List[ValueIdAndSelf] = list(map(ValueIdAndSelf, values.pop("planned", []))) if "planned" in values else None
    self.priority: dict = values.pop("priority", None)
    self.progress: Dict[str, int] = values.pop("progress", None)
    self.project: JiraProject = JiraProject(values=values.pop("project")) if "project" in values else None
    self.rank: str = values.pop("rank", None)
    self.reporter: JiraUser = JiraUser(values=values.pop("reporter")) if "reporter" in values else None
    self.resolution: IssueResolution = IssueResolution(values=values.pop("resolution")) if "resolution" in values else None
    self.resolutiondate: datetime = dates.parse_date(values.pop("resolutiondate")) if "resolutiondate" in values else None
    self.sprint: str = values.pop("sprint", None)
    self.sprintFinal: str = values.pop("sprintFinal", None)
    self.sprintRaw: List[str] = values.pop("sprintRaw", None)
    self.status: JiraStatus = JiraStatus(values=values.pop("status")) if "status" in values else None
    self.storyPoints: float = values.pop("storyPoints", None)
    self.subtasks: List[IssueBasic] = list(map(IssueBasic, values.pop("subtasks", []))) if "subtasks" in values else None
    self.summary: str = values.pop("summary", None)
    self.testCases: str = values.pop("testCases", None)
    self.thirdPartyType: ValueIdAndSelf = ValueIdAndSelf(values=values.pop("thirdPartyType")) if "thirdPartyType" in values else None
    self.timeestimate: int = values.pop("timeestimate", None)
    self.timeoriginalestimate: int = values.pop("timeoriginalestimate", None)
    self.timespent: int = values.pop("timespent", None)
    self.timetracking: dict = values.pop("timetracking", None)
    self.updated: datetime = dates.parse_iso_date(values.pop("updated")) if "updated" in values else None
    self.userType: ValueIdAndSelf = ValueIdAndSelf(values=values.pop("userType")) if "userType" in values else None
    self.votes: dict = values.pop("votes", None)
    self.watches: dict = values.pop("watches", None)
    self.worklog: dict = values.pop("worklog", None)
    self.workratio: int = values.pop("workratio", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class IssueBasic(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    if "fields" in values:
      values.update(values.pop("fields"))

    self.issuetype: IssueType = IssueType(values=values.pop("issuetype")) if "issuetype" in values else None
    self.key: str = values.pop("key", None)
    self.priority: dict = values.pop("priority", None)
    self.status: JiraStatus = JiraStatus(values=values.pop("status")) if "status" in values else None
    self.summary: str = values.pop("summary", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class IssueKeyAndValue(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.issueKey: str = values.pop("issueKey", None)
    self.value: float = float(values.pop("value")) if "value" in values else None

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class IssueLink(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.inwardIssue: IssueBasic = IssueBasic(values=values.pop("inwardIssue")) if "inwardIssue" in values else None
    self.outwardIssue: IssueBasic = IssueBasic(values=values.pop("outwardIssue")) if "outwardIssue" in values else None
    self.type: IssueLinkType = IssueLinkType(values=values.pop("type")) if "type" in values else None

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class IssueLinkType(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.inward: str = values.pop("inward", None)
    self.name: str = values.pop("name", None)
    self.outward: str = values.pop("outward", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class IssueResolution(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.description: str = values.pop("description", None)
    self.name: str = values.pop("name", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class IssueSearchResult(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.expand: str = values.pop("expand", None)
    self.issues: List[Issue] = list(map(Issue, values.pop("issues", []))) if "issues" in values else None
    self.maxResults: int = values.pop("maxResults", None)
    self.names: Dict[str, str] = values.pop("names", None)
    self.startAt: int = values.pop("startAt", None)
    self.total: int = values.pop("total", None)

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class IssueType(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.avatarId: int = values.pop("avatarId", None)
    self.description: str = values.pop("description", None)
    self.iconUrl: str = values.pop("iconUrl", None)
    self.name: str = values.pop("name", None)
    self.subtask: bool = values.pop("subtask", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class JiraProject(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.assigneeType: str = values.pop("assigneeType", None)
    self.avatarUrls: dict = values.pop("avatarUrls", None)
    self.description: str = values.pop("description", None)
    self.expand: str = values.pop("expand", None)
    self.isPrivate: str = values.pop("isPrivate", None)
    self.issueTypes: List[IssueType] = list(map(IssueType, values.pop("issueTypes", []))) if "issueTypes" in values else None
    self.key: str = values.pop("key", None)
    self.lead: JiraUser = JiraUser(values=values.pop("lead")) if "lead" in values else None
    self.name: str = values.pop("name", None)
    self.projectTypeKey: str = values.pop("projectTypeKey", None)
    self.properties: dict = values.pop("properties", None)
    self.roles: dict = values.pop("roles", None)
    self.simplified: bool = values.pop("simplified", None)
    self.style: str = values.pop("style", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class JiraStatus(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.description: str = values.pop("description", None)
    self.iconUrl: str = values.pop("iconUrl", None)
    self.name: str = values.pop("name", None)
    self.statusCategory: dict = values.pop("statusCategory", None)

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class JiraUser(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.active: bool = values.pop("active", None)
    self.avatarUrls: dict = values.pop("avatarUrls", None)
    self.displayName: str = values.pop("displayName", None)
    self.emailAddress: str = values.pop("emailAddress", None)
    self.key: str = values.pop("key", None)
    self.name: str = values.pop("name", None)
    self.self: str = values.pop("self", None)
    self.timeZone: str = values.pop("timeZone", None)

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class SprintState(enums.EnumAutoName):
  ACTIVE = auto()
  CLOSED = auto()
  FUTURE = auto()

  @staticmethod
  def from_string(state: str) -> Optional['SprintState']:
    return SprintState[state.upper()] if state else None


class Sprint(IdAndSelf, DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.canUpdateSprint: bool = values.pop("canUpdateSprint", None)
    self.completeDate: datetime = dates.parse_date(values.pop("completeDate")) if in_values_and_valid_date(values, "completeDate") else None
    self.daysRemaining: int = int(values.pop("daysRemaining")) if "daysRemaining" in values else None
    self.endDate: datetime = dates.parse_date(values.pop("endDate")) if in_values_and_valid_date(values, "endDate") else None
    self.goal: str = values.pop("goal", None)
    self.linkedPagesCount: int = int(values.pop("linkedPagesCount")) if "linkedPagesCount" in values else None
    self.name: str = values.pop("name", None)
    self.originBoardId: int = int(values.pop("originBoardId")) if "originBoardId" in values else None
    self.sequence: int = int(values.pop("sequence")) if "sequence" in values else None
    self.startDate: datetime = dates.parse_date(values.pop("startDate")) if in_values_and_valid_date(values, "startDate") else None
    self.state: SprintState = SprintState.from_string(values.pop("state")) if "state" in values else None

    IdAndSelf.__init__(self, values)
    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class SprintReportContents(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.allIssuesEstimateSum: TextAndValue = TextAndValue(values=values.pop("allIssuesEstimateSum")) if "allIssuesEstimateSum" in values else None
    self.completedIssues: List[dict] = values.pop("completedIssues", None)
    self.completedIssuesEstimateSum: TextAndValue = TextAndValue(values=values.pop("completedIssuesEstimateSum")) if "completedIssuesEstimateSum" in values else None
    self.completedIssuesInitialEstimateSum: TextAndValue = TextAndValue(
      values=values.pop("completedIssuesInitialEstimateSum")
    ) if "completedIssuesInitialEstimateSum" in values else None
    self.issueKeysAddedDuringSprint: Dict[str, bool] = values.pop("issueKeysAddedDuringSprint", None)
    self.issuesCompletedInAnotherSprint: List[dict] = values.pop("issuesCompletedInAnotherSprint", None)
    self.issuesCompletedInAnotherSprintEstimateSum: TextAndValue = TextAndValue(
      values=values.pop("issuesCompletedInAnotherSprintEstimateSum")
    ) if "issuesCompletedInAnotherSprintEstimateSum" in values else None
    self.issuesCompletedInAnotherSprintInitialEstimateSum: TextAndValue = TextAndValue(
      values=values.pop("issuesCompletedInAnotherSprintInitialEstimateSum")
    ) if "issuesCompletedInAnotherSprintInitialEstimateSum" in values else None
    self.issuesNotCompletedEstimateSum: TextAndValue = TextAndValue(values=values.pop("issuesNotCompletedEstimateSum")) if "issuesNotCompletedEstimateSum" in values else None
    self.issuesNotCompletedInCurrentSprint: List[dict] = values.pop("issuesNotCompletedInCurrentSprint", None)
    self.issuesNotCompletedInitialEstimateSum: TextAndValue = TextAndValue(
      values=values.pop("issuesNotCompletedInitialEstimateSum")
    ) if "issuesNotCompletedInitialEstimateSum" in values else None
    self.puntedIssues: List[str] = values.pop("puntedIssues", None)
    self.puntedIssuesEstimateSum: TextAndValue = TextAndValue(values=values.pop("puntedIssuesEstimateSum")) if "puntedIssuesEstimateSum" in values else None
    self.puntedIssuesInitialEstimateSum: TextAndValue = TextAndValue(values=values.pop("puntedIssuesInitialEstimateSum")) if "puntedIssuesInitialEstimateSum" in values else None

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class SprintReport(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.contents: SprintReportContents = SprintReportContents(values=values.pop("contents")) if "contents" in values else None
    self.lastUserToClose: str = values.pop("lastUserToClose", None)
    self.sprint: Sprint = Sprint(values=values.pop("sprint")) if "sprint" in values else None
    self.supportsPages: bool = values.pop("supportsPages", None)

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class VelocityReportStats(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.allConsideredIssueKeys: List[str] = values.pop("allConsideredIssueKeys", None)
    self.completed: TextAndValue = TextAndValue(values=values.pop("completed")) if "completed" in values else None
    self.completedEntries: List[IssueKeyAndValue] = list(map(IssueKeyAndValue, values.pop("completedEntries", []))) if "completedEntries" in values else None
    self.estimated: TextAndValue = TextAndValue(values=values.pop("estimated")) if "estimated" in values else None
    self.estimatedEntries: List[IssueKeyAndValue] = list(map(IssueKeyAndValue, values.pop("estimatedEntries", []))) if "estimatedEntries" in values else None

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class VelocityReport(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.sprints: List[Sprint] = list(map(Sprint, values.pop("sprints", []))) if "sprints" in values else None
    self.velocityStatEntries: Dict[str, VelocityReportStats] = dict((str(k), VelocityReportStats(values=v)) for (k, v) in values.pop("velocityStatEntries").items()
                                                                   ) if "velocityStatEntries" in values else None

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)



#!/usr/bin/env python
import argparse
import logging
import os
import shutil
import subprocess
import sys
import threading
from pathlib import Path
from typing import Union

from ltpylib import opts

LOG_FORMAT_PART_LEVEL = "{levelname:<8}"
LOG_FORMAT_PART_MESSAGE = "{message}"
LOG_FORMAT_PART_TIMESTAMP = "{asctime}"

DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_LOG_FORMAT = f"{LOG_FORMAT_PART_MESSAGE}"
DEFAULT_LOG_STYLE = "{"
LOG_SEP = "----------------------------------------------------------------------------------------------------------------------------------"

LOG_FORMAT_WITH_LEVEL = f"{LOG_FORMAT_PART_LEVEL} {LOG_FORMAT_PART_MESSAGE}"
LOG_FORMAT_WITH_TIMESTAMP = f"[{LOG_FORMAT_PART_TIMESTAMP}] {LOG_FORMAT_PART_MESSAGE}"


# see https://stackoverflow.com/questions/21953835/run-subprocess-and-print-output-to-logging
class LogPipe(threading.Thread):

  def __init__(self, level: int = logging.INFO):
    threading.Thread.__init__(self)
    self.daemon = False
    self.level = level
    self.fd_read, self.fd_write = os.pipe()
    self.pipe_reader = os.fdopen(self.fd_read)
    self.start()

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, exc_tb):
    self.close()

  def fileno(self):
    return self.fd_write

  def run(self):
    for line in iter(self.pipe_reader.readline, ''):
      logging.log(self.level, line.strip('\n'))

    self.pipe_reader.close()

  def close(self):
    os.close(self.fd_write)

  def write(self, message):
    logging.log(self.level, message)

  def flush(self):
    pass


class StderrStreamHandler(logging.StreamHandler):

  def __init__(self):
    super().__init__(stream=sys.stderr)

  def handleError(self, record):
    err_type, err_value, traceback = sys.exc_info()
    if err_type == BrokenPipeError:
      exit(0)

    super().handleError(record)


class StdoutStreamHandler(logging.StreamHandler):

  def __init__(self):
    super().__init__(stream=sys.stdout)

  def handleError(self, record):
    err_type, err_value, traceback = sys.exc_info()
    if err_type == BrokenPipeError:
      exit(0)

    super().handleError(record)


def init_logging(
  verbose: bool = None,
  quiet: bool = None,
  log_level: Union[int, str] = None,
  log_format: str = None,
  args: Union[argparse.Namespace, opts.BaseArgs] = None,
  use_stderr: bool = False,
):
  if args:
    if verbose is None and hasattr(args, "verbose"):
      verbose = args.verbose
    if log_level is None and hasattr(args, "log_level"):
      log_level = args.log_level
    if log_format is None and hasattr(args, "log_format"):
      log_format = args.log_format
    if quiet is None and hasattr(args, "quiet"):
      quiet = args.quiet

  if verbose:
    log_level = logging.DEBUG
  elif quiet:
    log_level = logging.WARNING
  elif log_level is not None:
    log_level = log_level
  elif os.environ.get("log_level") is not None:
    log_level = os.environ.get("log_level")
  else:
    log_level = DEFAULT_LOG_LEVEL

  log_config_kwargs = {
    "style": DEFAULT_LOG_STYLE,
    "handlers": [StderrStreamHandler() if use_stderr else StdoutStreamHandler()],
  }
  logging.basicConfig(
    level=log_level,
    format=log_format if log_format else DEFAULT_LOG_FORMAT,
    **log_config_kwargs,
  )


def add_file_logging(log_file: Path):
  if not log_file.parent.exists():
    log_file.parent.mkdir(parents=True)

  root_logger = logging.getLogger()

  file_handler = logging.FileHandler(log_file.as_posix())
  file_handler.setFormatter(root_logger.handlers[0].formatter)

  root_logger.addHandler(file_handler)


def is_debug_enabled():
  return logging.root.isEnabledFor(logging.DEBUG)


def log_sep(debug_only=False):
  if debug_only:
    logging.debug(LOG_SEP)
  else:
    logging.info(LOG_SEP)


def create_path_log_info(path: Path, replace_home_dir: bool = True) -> str:
  output = path.as_posix()

  if replace_home_dir:
    output = output.replace(os.getenv('HOME'), '~', 1)

  return output


def log_with_sep(msg, *args, level: int = logging.INFO, **kwargs):
  logging.log(level, LOG_SEP)
  logging.log(level, msg, *args, **kwargs)
  logging.log(level, LOG_SEP)


def log_with_title_sep(title, *args, msg=None, level: int = logging.INFO, **kwargs):
  logging.log(level, title)
  logging.log(level, LOG_SEP)
  if msg is not None:
    logging.log(level, msg, *args, **kwargs)
    logging.log(level, '')


def log_title_with_sep(title, level: int = logging.INFO):
  logging.log(level, title)
  logging.log(level, LOG_SEP)


def ltlogs_dir() -> Path:
  return Path(os.getenv("LTLOGS_DIR", os.path.expanduser("~/Library/Logs/lt_logs")))


def tail_log_file(file: str, *func_args):
  if shutil.which('multitail'):
    log_cmd = ['multitail']
    mt_conf = Path(os.getenv('DOTFILES') + '/multitail.conf')
    if '-F' not in func_args and mt_conf.is_file():
      log_cmd.extend(['-F', mt_conf.as_posix()])

    if '-CS' not in func_args:
      log_cmd.extend(['-CS', 'l4j'])

    if '-n' not in func_args:
      log_cmd.extend(['-n', '1500'])

  else:
    log_cmd = ['tail', '-1500f']

  if func_args:
    log_cmd.extend(func_args)

  log_cmd.append(file)

  subprocess.check_call(log_cmd, universal_newlines=True)


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
import unittest

from ltpylib.mvnutil import MavenArtifact

has_all = MavenArtifact(
  group_id="com.lancethomps",
  artifact_id="lava",
  version="1.0.0",
  packaging="jar",
  classifier="shaded",
  scope="runtime",
)
has_group_artifact = MavenArtifact(**{k: getattr(has_all, k) for k in ["group_id", "artifact_id"]})
has_artifact = MavenArtifact(**{k: getattr(has_all, k) for k in ["artifact_id"]})
has_group_artifact_classifier = MavenArtifact(**{k: getattr(has_all, k) for k in ["group_id", "artifact_id", "classifier"]})
artifacts = [
  has_all,
  has_group_artifact,
  has_artifact,
  has_group_artifact_classifier,
]


class TestMavenArtifact(unittest.TestCase):

  def test_to_artifact_string(self):
    assert has_all.to_artifact_string() == "com.lancethomps:lava:1.0.0:jar:shaded:runtime"
    assert has_group_artifact.to_artifact_string() == "com.lancethomps:lava"
    assert has_artifact.to_artifact_string() == "*:lava"
    assert has_group_artifact_classifier.to_artifact_string() == "com.lancethomps:lava:*:*:shaded"

  def test_from_artifact_string(self):
    for artifact in artifacts:
      assert artifact == MavenArtifact.from_artifact_string(artifact.to_artifact_string(), strict=False)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python

import logging
import re
from pathlib import Path
from typing import Callable, List, Match, Pattern, Sequence, Union

from ltpylib import files, logs

SKIP_FILE_REGEX: Pattern = re.compile(r"^[^a-zA-Z0-9]*template:skip\s*$")
START_REGEX = re.compile(r"^[^a-zA-Z0-9]*template:start (.*?)$")
END_REGEX = re.compile(r"^[^a-zA-Z0-9]*template:end\s*$")
DEFAULT_CANDIDATE_FILES_GLOBS: Sequence[str] = ('**/*',)
DEFAULT_TEMPLATE_FILES_GLOBS: Sequence[str] = ('**/*.template.*',)


def replace_templates_in_files(
  candidate_files_dir: Path,
  template_dir: Path,
  candidate_files_globs: List[str] = DEFAULT_CANDIDATE_FILES_GLOBS,
  template_files_globs: List[str] = DEFAULT_TEMPLATE_FILES_GLOBS,
  debug_mode: bool = False,
  force_replace: bool = False,
  match_leading_whitespace: bool = False,
) -> bool:
  candidate_files: List[Path] = files.filter_files_with_matching_line(
    files.list_files(candidate_files_dir, candidate_files_globs),
    [SKIP_FILE_REGEX],
    check_n_lines=2,
  )
  if not candidate_files:
    logging.warning("No candidate files found: candidate_files_dir=%s candidate_files_globs=%s", candidate_files_dir.as_posix(), candidate_files_globs)
    return False

  logs.log_with_title_sep('Candidate Files', '\n'.join([file.absolute().as_posix() for file in candidate_files]), level=logging.DEBUG)

  template_files: List[Path] = files.list_files(template_dir, template_files_globs)
  if not template_files:
    logging.warning("No template files found: template_dir=%s template_files_globs=%s", template_dir.as_posix(), template_files_globs)
    return False

  logs.log_with_title_sep('Template Files', '\n'.join([file.absolute().as_posix() for file in template_files]), level=logging.DEBUG)

  for template_file in template_files:
    replace_template(
      candidate_files,
      template_file,
      debug_mode=debug_mode,
      force_replace=force_replace,
      match_leading_whitespace=match_leading_whitespace,
    )

  return True


def replace_template(
  candidate_files: List[Path],
  template_file: Path,
  debug_mode: bool = False,
  force_replace: bool = False,
  match_leading_whitespace: bool = False,
):
  template_file_content: str = files.read_file(template_file)

  lines = []
  found_start = False
  search_string: str = ''

  for line in template_file_content.splitlines():
    if found_start:
      if END_REGEX.fullmatch(line):
        break

      lines.append(line)
    elif START_REGEX.fullmatch(line):
      found_start = True
      search_string = START_REGEX.fullmatch(line).group(1)

  if not lines or not search_string:
    logging.error('Something went wrong, skipping template: template_file=%s lines_count=%s search_string=%s', template_file.as_posix(), len(lines), search_string)
    return

  replacement_str: str = '\n'.join(lines)
  replacement: Union[str, Callable[[Match], str]] = replacement_str

  search_string_groups = re.compile(search_string).groups
  if search_string_groups > 0:

    def repl(match: Match) -> str:
      groups = match.groupdict()
      repl_string = ""
      if "start" in groups:
        repl_string += groups.get("start")

      repl_string += replacement_str

      if "end" in groups:
        repl_string += groups.get("end")

      return repl_string

    replacement = repl
  elif match_leading_whitespace:
    replacement_leading_spaces = len(lines[0]) - len(lines[0].lstrip())

    def repl(match: re.Match) -> str:
      matched_string = match.group()
      first_line = matched_string.splitlines()[0]
      match_leading_spaces = len(first_line) - len(first_line.lstrip())
      if match_leading_spaces != replacement_leading_spaces:
        replacement_spaces_regex = "^ {" + str(replacement_leading_spaces) + "}"
        replacement_spaces_replacement = " " * match_leading_spaces
        return re.sub(replacement_spaces_regex, replacement_spaces_replacement, replacement_str, flags=re.MULTILINE)

      return replacement_str

    replacement = repl

  logs.log_with_title_sep('template_file', template_file.as_posix(), level=logging.DEBUG)
  logs.log_with_title_sep('search_string', search_string, level=logging.DEBUG)
  logs.log_with_title_sep('replacement_str', replacement_str, level=logging.DEBUG)

  if not debug_mode:
    for file in candidate_files:
      if file.absolute().as_posix() == template_file.absolute().as_posix():
        continue

      if files.replace_matches_in_file(file, search_string, replacement, quote_replacement=False, force_replace=force_replace):
        logging.info("Replaced template: %s -> %s", template_file.as_posix(), file.as_posix())


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
# pylint: disable=C0111
from typing import Any, Callable, Dict, List, Optional, Sequence, TypeVar, Union

from ltpylib import checks, strings
from ltpylib.collect import modify_list_of_dicts

T = TypeVar('T')


def convert_keys_to_snake_case(
  obj: Union[dict, list],
  recursive: bool = False,
) -> Union[dict, list]:
  if isinstance(obj, list):
    objs = obj
  else:
    objs = [obj]

  for obj_dict in objs:
    dict_items = list(obj_dict.items())
    for key, val in dict_items:
      key_snake_case = strings.to_snake_case(key)
      if key != key_snake_case:
        obj_dict[key_snake_case] = obj_dict.pop(key)

      if recursive and isinstance(val, dict):
        convert_keys_to_snake_case(
          val,
          recursive=recursive,
        )
      elif recursive and isinstance(val, list) and len(val) > 0 and isinstance(val[0], dict):
        for inner_val in val:
          convert_keys_to_snake_case(
            inner_val,
            recursive=recursive,
          )

  return obj


def convert_boolean_values_to_string(
  value_to_convert: Union[dict, list],
  recursive: bool = True,
  ignore_fields: List[str] = None,
  only_fields: List[str] = None,
) -> Union[dict, list]:
  if isinstance(value_to_convert, list):
    if isinstance(value_to_convert[0], str):
      return value_to_convert

    objs = value_to_convert
  else:
    objs = [value_to_convert]

  for obj_dict in objs:
    items = obj_dict.items() if not only_fields else [(f, obj_dict[f]) for f in only_fields if f in obj_dict]
    for key, val in items:
      if isinstance(val, bool):
        if ignore_fields and key in ignore_fields:
          continue

        obj_dict[key] = str(val).lower()
      elif recursive and isinstance(val, dict):
        convert_boolean_values_to_string(
          val,
          recursive=recursive,
          ignore_fields=ignore_fields,
          only_fields=only_fields,
        )
      elif recursive and isinstance(val, list) and len(val) > 0 and isinstance(val[0], dict):
        for inner_val in val:
          convert_boolean_values_to_string(
            inner_val,
            recursive=recursive,
            ignore_fields=ignore_fields,
            only_fields=only_fields,
          )

  return value_to_convert


def convert_string_values_to_correct_type(
  value_to_convert: Union[dict, list],
  convert_numbers: bool = True,
  convert_booleans: bool = True,
  use_decimal: bool = False,
  recursive: bool = False,
  ignore_fields: List[str] = None,
) -> Union[dict, list]:
  if isinstance(value_to_convert, list):
    if isinstance(value_to_convert[0], str):
      return [convert_string_to_correct_type(val, convert_numbers=convert_numbers, convert_booleans=convert_booleans, use_decimal=use_decimal) for val in value_to_convert]

    objs = value_to_convert
  else:
    objs = [value_to_convert]

  for obj_dict in objs:
    for key, val in obj_dict.items():
      if isinstance(val, str):
        if ignore_fields and key in ignore_fields:
          continue

        obj_dict[key] = convert_string_to_correct_type(val, convert_numbers=convert_numbers, convert_booleans=convert_booleans, use_decimal=use_decimal)
      elif recursive and isinstance(val, dict):
        convert_string_values_to_correct_type(
          val,
          convert_numbers=convert_numbers,
          convert_booleans=convert_booleans,
          use_decimal=use_decimal,
          recursive=recursive,
          ignore_fields=ignore_fields,
        )
      elif recursive and isinstance(val, list) and len(val) > 0 and isinstance(val[0], dict):
        for inner_val in val:
          convert_string_values_to_correct_type(
            inner_val,
            convert_numbers=convert_numbers,
            convert_booleans=convert_booleans,
            use_decimal=use_decimal,
            recursive=recursive,
            ignore_fields=ignore_fields,
          )

  return value_to_convert


def convert_string_to_correct_type(
  val: str,
  convert_numbers: bool = True,
  convert_booleans: bool = True,
  use_decimal: bool = False,
):
  if convert_numbers and strings.is_number(val, allow_comma=True):
    return strings.convert_to_number(val, use_decimal=use_decimal, remove_commas=True)
  elif convert_booleans and strings.is_boolean(val):
    return strings.convert_to_bool(val)

  return val


def copy_fields(
  from_val: dict,
  to_val: dict,
  fields: List[str],
  field_converter: Callable[[str], str] = None,
  field_converter_map: Dict[str, str] = None,
) -> dict:
  if from_val:
    for field in fields:
      if field in from_val:
        if field_converter is not None:
          to_val[field_converter(field)] = from_val[field]
        elif field_converter_map:
          to_val[field_converter_map.get(field, field)] = from_val[field]
        else:
          to_val[field] = from_val[field]

  return to_val


def find(key: str, obj: dict, yield_parent: bool = False) -> List[dict]:
  if isinstance(obj, dict):
    for k, v in list(obj.items()):
      if k == key:
        if yield_parent:
          yield obj
        else:
          yield v
      else:
        for res in find(key, v, yield_parent):
          yield res
  elif isinstance(obj, list):
    for d in obj:
      for res in find(key, d, yield_parent):
        yield res


def create_key_getter(key: Union[str, Callable[[T], Any]], is_dict: bool = True) -> Callable[[T], Any]:
  if isinstance(key, str):

    if is_dict:

      def key_getter(x):
        return x.get(key)
    else:

      def key_getter(x):
        return getattr(x, key)

  else:
    key_getter = key

  return key_getter


def find_first_with_key_value(list_of_dicts: List[dict], key: Union[str, Callable[[dict], Any]], expected_value: Any) -> Optional[dict]:
  key_getter = create_key_getter(key)

  for val in list_of_dicts:
    field_value = key_getter(val)
    if field_value == expected_value:
      return val


def group_by(list_of_dicts: List[T], key: Union[str, Callable[[T], Any]], is_dict: bool = True) -> Dict[Any, List[T]]:
  key_getter = create_key_getter(key, is_dict=is_dict)
  by_field: Dict[str, List[T]] = {}
  for val in list_of_dicts:
    field_value = key_getter(val)
    if field_value not in by_field:
      by_field[field_value] = []

    by_field[field_value].append(val)

  return by_field


def modify_dict_fields(
  datas: List[dict],
  fields_included: Sequence[str] = None,
  fields_order: Sequence[str] = None,
  in_place: bool = False,
) -> List[dict]:
  result: List[dict] = datas

  if fields_included:

    def fn(updated_data: dict):
      for field in list(updated_data.keys()):
        if field not in fields_included:
          updated_data.pop(field)

    result = modify_list_of_dicts(result, fn, in_place=in_place)

  if fields_order:
    if in_place:
      data_copy = result[0].copy()
      ordered_data = result[0]
      ordered_data.clear()
    else:
      data_copy = result[0].copy()
      ordered_data = dict()

    for field in fields_order:
      ordered_data[field] = data_copy.pop(field, None)

    ordered_data.update(data_copy)
    result[0] = ordered_data

  return result


def prefix_dict_keys(require_prefix: str, data: Dict[str, Any], always_add: str = None) -> dict:
  updated_data = {}
  for input_key, value in data.items():
    key = input_key
    if not key.startswith(require_prefix):
      key = require_prefix + key

    if always_add:
      key = always_add + key

    updated_data[key] = value

  return updated_data


def unique_key_values(list_of_dicts: List[dict], key: Union[str, Callable[[dict], Any]], include_nulls: bool = False) -> List[Any]:
  key_getter = create_key_getter(key)
  unique_values = []
  for val in list_of_dicts:
    field_value = key_getter(val)
    if field_value is None and not include_nulls:
      continue

    if field_value not in unique_values:
      unique_values.append(field_value)

  return unique_values


def remove_nulls(dict_with_nulls: dict) -> dict:
  return {key: val for (key, val) in dict_with_nulls.items() if val is not None}


def remove_nulls_and_empty(dict_with_nulls: dict) -> dict:
  return {key: val for (key, val) in dict_with_nulls.items() if checks.is_not_empty(val)}


if __name__ == "__main__":
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)



#!/usr/bin/env python

APPEND = "append"
APPEND_CONST = "append_const"
COUNT = "count"
EXTEND = "extend"
HELP = "help"
PARSERS = "parsers"
STORE = "store"
STORE_CONST = "store_const"
STORE_FALSE = "store_false"
STORE_TRUE = "store_true"
VERSION = "version"



#!/usr/bin/env python
import re
from pathlib import Path
from typing import List, Union
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

import xmltodict

from ltpylib import files, procs


def xq(file: Union[str, Path], expr: str, remove_null: bool = True, raw_output: bool = True) -> str:
  if isinstance(file, str):
    file = Path(file)

  xq_command: List[str] = ['xq']

  if remove_null:
    expr += ' | select (.!=null)'

  if raw_output:
    xq_command.append('--raw-output')

  xq_command.extend([expr, file.as_posix()])

  result: str = procs.run_and_parse_output(xq_command, check=True)[1].strip()

  if result == "":
    return None

  return result


def parse_xml_to_dict(file: Union[str, Path]) -> dict:
  if isinstance(file, str):
    file = Path(file)

  with open(file.as_posix()) as fd:
    return xmltodict.parse(fd.read())


def parse_xml(file: Union[str, Path], remove_namespace: bool = True) -> Element:
  if isinstance(file, str):
    file = Path(file)

  if remove_namespace:
    xml_string: str = re.sub(' xmlns="[^"]+"', '', files.read_file(file), count=1)
    return ElementTree.fromstring(xml_string)

  return ElementTree.parse(file.as_posix()).getroot()


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
import os
import unittest
from pathlib import Path

from ltpylib import strings
from ltpylibtests import testing_utils


class TestStrings(unittest.TestCase):

  def test_to_snake_case(self):
    test_file = Path(os.path.dirname(os.path.realpath(__file__))).joinpath("resources/test_to_snake_case.yaml")
    config = testing_utils.load_yaml_test_config_file(test_file)

    assert len(config.test_cases) > 0

    for test_case in config.test_cases:
      if not test_case:
        continue

      self.assertEqual(strings.to_snake_case(test_case.input), test_case.expected)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
# pylint: disable=C0111
import argparse
import inspect
import logging
import psutil
import shutil
import subprocess
from pathlib import Path
from typing import List, Tuple, Union

from ltpylib import files, logs, opts, procs


class ProcessRunner(object):

  def __init__(
    self,
    proc_id: str,
    start_cmd: List[str],
    build_cmds: List[Union[List[str], Tuple[Union[str, List[str]], dict]]] = None,
    force_build_before_start=None,
    log_file: str = None,
    log_multitail_args: List[str] = None,
    pid_file: str = None,
    proc_name_matcher: str = None,
    status_loop_wait_time: int = 30,
    status_loop_sleep_time: int = 1
  ):
    """
    :type force_build_before_start: Callable[[ProcessRunner], bool]
    """
    self.proc_id = proc_id
    self.start_cmd = start_cmd
    self.build_cmds = build_cmds
    self.force_build_before_start = force_build_before_start
    self.log_file = log_file or logs.ltlogs_dir().joinpath(f"{proc_id}.log").as_posix()
    self.log_multitail_args = log_multitail_args
    self.pid_file = pid_file
    self.proc_name_matcher = proc_name_matcher
    self.status_loop_wait_time = status_loop_wait_time
    self.status_loop_sleep_time = status_loop_sleep_time

  def get_proc_pid(self) -> Union[int, None]:
    if self.proc_name_matcher:
      matched_procs = procs.get_procs_from_name(self.proc_name_matcher)
      if matched_procs:
        return matched_procs[0][0]

      return None
    elif self.pid_file:
      if not Path(self.pid_file).is_file():
        return None

      pid = int(files.read_file(self.pid_file))
      if psutil.pid_exists(pid):
        return pid

      return None
    else:
      raise Exception('No pid_file or proc_name_matcher specified.')

  def check_proc_running(self):
    return self.get_proc_pid() is not None

  def log(self):
    if self.log_multitail_args and shutil.which('multitail'):
      logs.tail_log_file(self.log_file, *self.log_multitail_args)
    else:
      logs.tail_log_file(self.log_file)

  def build(self):
    if not self.build_cmds:
      raise Exception('No build_cmds specified.')

    for build_cmd_opt in self.build_cmds:
      func_args = {'universal_newlines': True}
      if isinstance(build_cmd_opt, tuple):
        build_cmd = build_cmd_opt[0]
        func_args.update(build_cmd_opt[1])
      else:
        build_cmd = build_cmd_opt

      subprocess.check_call(build_cmd, **func_args)

  def _await_termination(self, pid: int = None):
    if pid is None:
      pid = self.get_proc_pid()

    procs.await_termination(
      pid,
      timeout=self.status_loop_wait_time,
      sleep_time=self.status_loop_sleep_time,
      log_level=logging.INFO,
    )

  def stop(self) -> bool:
    pid = self.get_proc_pid()
    stop_result = False
    if pid is not None:
      stop_result = procs.stop_proc_by_pid(pid)
      if stop_result:
        self._await_termination(pid)

    return stop_result

  def start(self) -> int:
    self.stop()

    if self.force_build_before_start is not None and self.force_build_before_start(self):
      logging.info('Build required before executing start command - building now...')
      self.build()

    log_file_fd = open(self.log_file, 'a')
    pid = subprocess.Popen(
      self.start_cmd,
      stdout=log_file_fd,
      stderr=log_file_fd,
      universal_newlines=True,
    ).pid

    if pid is not None and self.pid_file:
      files.write_file(self.pid_file, str(pid))

    self.status(pid)
    return pid

  def status(self, pid: int = None) -> bool:
    if pid is None:
      pid = self.get_proc_pid()

    if pid is not None:
      proc = psutil.Process(pid)
      if proc.is_running():
        logging.info('%s', procs.proc_debug_string(proc))
        return True

    return False

  def _exec(self, cmd: str):
    return getattr(self, cmd)()

  def _parse_args(self) -> argparse.Namespace:
    arg_parser = opts.create_default_arg_parser()
    choices = []
    skip_methods = ['parse_and_run', 'check_proc_running']
    methods = inspect.getmembers(self, predicate=inspect.ismethod)
    for method in methods:
      if not method[0].startswith('_') and method[0] not in skip_methods:
        choices.append(method[0])
    arg_parser.add_argument('actions', choices=choices, nargs='+')
    return opts.parse_args_and_init_others(arg_parser)

  def parse_and_run(self):
    args = self._parse_args()
    result = None
    for cmd in args.actions:
      result = self._exec(cmd)

    if isinstance(result, bool):
      if not result:
        exit(1)
      else:
        exit(0)
    elif result is not None:
      logging.info('%s', result)



#!/usr/bin/env python
import argparse
import os
from pathlib import Path
from typing import Any, Callable, List, Optional, Sequence, Union

from ltpylib import opts_actions
from ltpylib.common_types import TypeWithDictRepr
from ltpylib.opts_actions import APPEND, STORE_TRUE
from ltpylib.output import is_output_to_terminal

TRUE_VALUES = ["true", "1", "t", "yes", "y"]
DEFAULT_POSITIONALS_KEY = "command"


class PositionalsHelpFormatter(argparse.HelpFormatter):

  def __init__(
    self,
    prog,
    indent_increment=2,
    max_help_position=24,
    width=None,
    positionals_key: str = DEFAULT_POSITIONALS_KEY,
    positionals_type: Callable = None,
  ):
    super().__init__(prog, indent_increment, max_help_position, width)
    self.positionals_key = positionals_key
    self.positionals_type = positionals_type
    self.positionals_action = argparse._StoreAction([], self.positionals_key, nargs="+")

  def add_usage(self, usage, actions, groups, prefix=None):
    from itertools import chain

    super().add_usage(usage, chain(actions, [self.positionals_action]), groups, prefix)

  def add_arguments(self, actions):
    from itertools import chain

    if self._current_section.heading == "positional arguments":
      super().add_arguments(chain(actions, [self.positionals_action]))
    else:
      super().add_arguments(actions)


class ActionPathValidate(argparse.Action):

  def __init__(self, option_strings, dest, type=None, **kwargs):
    if type is not None:
      raise ValueError("type not allowed")

    super(ActionPathValidate, self).__init__(option_strings=option_strings, dest=dest, type=Path, **kwargs)

  def __call__(self, parser, namespace, value: Path, option_string=None):
    setattr(namespace, self.dest, value)

    if value is not None:
      if not value.exists():
        raise ValueError("Supplied Path does not exist: %s" % (value.as_posix()))


class ActionPathsValidate(argparse._AppendAction):

  def __init__(self, option_strings, dest, type=None, **kwargs):
    if type is not None:
      raise ValueError("type not allowed")

    super(ActionPathsValidate, self).__init__(option_strings=option_strings, dest=dest, type=Path, **kwargs)

  def __call__(self, parser, namespace, values: List[Path], option_string=None):
    setattr(namespace, self.dest, values)

    if values:
      for item in values:
        if not item.exists():
          raise ValueError("Supplied Path does not exist: %s" % (item.as_posix()))


class BaseArgs(TypeWithDictRepr):

  def __init__(self, args: argparse.Namespace):
    self._args: argparse.Namespace = args
    self.debug: bool = args.debug
    self.verbose: bool = args.verbose

  def check_for_debug_and_exit(self, exit_code: int = 0, log_args_before_exit: bool = False):
    if self.debug:
      import logging

      if log_args_before_exit:
        log_args(self)

      logging.warning("debug enabled, exiting...")
      exit(exit_code)


class ColorArgs(object):

  def __init__(self, args: argparse.Namespace):
    self.color: bool = args.color if args.color is not None else is_output_to_terminal()

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser, default: bool = None) -> argparse.ArgumentParser:
    arg_parser.add_argument("--color", action=argparse.BooleanOptionalAction, default=default)
    return arg_parser


class IncludeExcludeCmdArgs(object):

  def __init__(self, args: argparse.Namespace):
    self.exclude_commands: List[str] = args.exclude_cmd or []
    self.include_commands: List[str] = args.include_cmd or []

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    arg_parser.add_argument("--exclude-cmd", action=APPEND)
    arg_parser.add_argument("--include-cmd", action=APPEND)
    return arg_parser


class IncludeExcludeRegexArgs(object):

  def __init__(self, args: argparse.Namespace):
    self.exclude_regex: List[str] = args.exclude_regex or []
    self.include_regex: List[str] = args.include_regex or []

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    arg_parser.add_argument("--exclude-regex", action=APPEND)
    arg_parser.add_argument("--include-regex", action=APPEND)
    return arg_parser


class LoggingArgs(object):

  def __init__(self, args: argparse.Namespace):
    self.log_format: str = args.log_format
    self.log_level: str = args.log_level
    self.quiet: bool = args.quiet

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    arg_parser.add_argument("--log-format")
    arg_parser.add_argument("--log-level")
    arg_parser.add_argument("--quiet", action=STORE_TRUE)
    return arg_parser


class OutputFormatArgs(object):

  def __init__(self, args: argparse.Namespace):
    self.json: bool = args.json
    self.markdown: bool = args.markdown
    self.yaml: bool = args.yaml

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    group = arg_parser.add_mutually_exclusive_group()
    group.add_argument("--json", action=opts_actions.STORE_TRUE)
    group.add_argument("--markdown", action=opts_actions.STORE_TRUE)
    group.add_argument("--yaml", action=opts_actions.STORE_TRUE)
    return arg_parser


class PagerArgs(object):

  def __init__(self, args: argparse.Namespace):
    self.no_pager: bool = args.no_pager
    self.pager: str = args.pager
    self.use_pager: bool = args.use_pager

  def should_use_pager(self, default: bool = True) -> bool:
    if self.use_pager:
      return True

    if self.no_pager:
      return False

    return default and is_output_to_terminal()

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser, default_pager: str = None) -> argparse.ArgumentParser:
    arg_parser.add_argument("--no-pager", action=STORE_TRUE)
    arg_parser.add_argument("--pager", default=default_pager if default_pager else os.getenv("PAGER", "less"))
    arg_parser.add_argument("--use-pager", action=STORE_TRUE)
    return arg_parser


class RegexCasingArgs(object):

  def __init__(self, args: argparse.Namespace):
    self.ignore_case: bool = args.ignore_case
    self.use_case: bool = args.use_case

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    arg_parser.add_argument("--ignore-case", "-i", action=STORE_TRUE)
    arg_parser.add_argument("--use-case", action=STORE_TRUE)
    return arg_parser

  def should_ignore_case(self, regex_pattern: str) -> bool:
    if self.use_case:
      return False

    if self.ignore_case:
      return True

    return regex_pattern is not None and regex_pattern.islower()


def add_default_arguments_to_parser(arg_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
  arg_parser.add_argument("-v", "--verbose", action=STORE_TRUE)
  arg_parser.add_argument("--debug", action=STORE_TRUE)
  return arg_parser


def check_debug_mode() -> bool:
  return os.getenv("debug_mode", "false").lower() in TRUE_VALUES


def check_verbose() -> bool:
  return os.getenv("verbose", "false").lower() in TRUE_VALUES


def config_home_dir() -> Path:
  return Path(os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config")))


def create_default_arg_parser(
  has_positionals: bool = False,
  positionals_key: str = DEFAULT_POSITIONALS_KEY,
  positionals_type: Callable = None,
) -> argparse.ArgumentParser:
  if has_positionals:
    return create_default_with_positionals_arg_parser(positionals_key=positionals_key, positionals_type=positionals_type)

  arg_parser = argparse.ArgumentParser()
  return add_default_arguments_to_parser(arg_parser)


def create_default_with_positionals_arg_parser(
  positionals_key: str = DEFAULT_POSITIONALS_KEY,
  positionals_type: Callable = None,
) -> argparse.ArgumentParser:
  import functools

  arg_parser = argparse.ArgumentParser(formatter_class=functools.partial(PositionalsHelpFormatter, positionals_key=positionals_key, positionals_type=positionals_type))
  return add_default_arguments_to_parser(arg_parser)


def parse_args(arg_parser: argparse.ArgumentParser, argv: Optional[Sequence[str]] = None) -> argparse.Namespace:
  import argcomplete

  argcomplete.autocomplete(arg_parser)
  args = arg_parser.parse_args(args=argv)
  return args


def parse_args_and_init_others(
  arg_parser: argparse.ArgumentParser,
  argv: Optional[Sequence[str]] = None,
  log_level: Union[str, int] = None,
  log_format: str = None,
  has_positionals: bool = False,
  log_use_stderr: bool = False,
) -> argparse.Namespace:
  if has_positionals:
    return parse_args_with_positionals_and_init_others(arg_parser, argv=argv, log_level=log_level, log_format=log_format, log_use_stderr=log_use_stderr)

  from ltpylib.logs import init_logging

  if has_positionals_formatter_class(arg_parser):
    return parse_args_with_positionals_and_init_others(arg_parser, argv=argv, log_level=log_level, log_format=log_format, log_use_stderr=log_use_stderr)

  args = parse_args(arg_parser, argv=argv)
  init_logging(args=args, log_level=log_level, log_format=log_format, use_stderr=log_use_stderr)
  return args


def parse_args_with_positionals_and_init_others(
  arg_parser: argparse.ArgumentParser,
  argv: Optional[Sequence[str]] = None,
  log_level: Union[str, int] = None,
  log_format: str = None,
  log_use_stderr: bool = False,
  positionals_key: str = None,
  positionals_type: Callable = None,
) -> argparse.Namespace:
  import argcomplete
  from ltpylib.logs import init_logging

  argcomplete.autocomplete(arg_parser)
  args, positionals = arg_parser.parse_known_intermixed_args(args=argv)

  if positionals_key is None:
    if has_positionals_formatter_class(arg_parser):
      positionals_key = get_positionals_formatter_field(arg_parser, "positionals_key")
    else:
      positionals_key = DEFAULT_POSITIONALS_KEY

  if positionals_type is None and has_positionals_formatter_class(arg_parser):
    positionals_type = get_positionals_formatter_field(arg_parser, "positionals_type")

  if positionals_type is not None:
    positionals = [positionals_type(arg) for arg in positionals]

  args.__setattr__(positionals_key, positionals)
  init_logging(args=args, log_level=log_level, log_format=log_format, use_stderr=log_use_stderr)
  return args


def get_positionals_formatter_field(arg_parser: argparse.ArgumentParser, field: str) -> Any:
  import functools

  help_formatter: functools.partial = arg_parser.formatter_class
  return help_formatter.keywords.get(field)


def has_positionals_formatter_class(arg_parser: argparse.ArgumentParser) -> bool:
  import functools

  fc = arg_parser.formatter_class
  return isinstance(fc, functools.partial) and fc.func == PositionalsHelpFormatter


def does_stdin_have_data() -> bool:
  import sys
  import select

  if select.select([sys.stdin], [], [], 0.0)[0]:
    return True
  elif sys.stdin.isatty():
    return True
  else:
    return False


def log_args(args: BaseArgs, include_raw_args: bool = True, only_keys: List[str] = None, skip_keys: List[str] = None):
  import logging
  from ltpylib.logs import log_sep

  logging.info("ARGS")
  log_sep(debug_only=False)

  log_format = "%-" + str(len(max(args.__dict__.keys(), key=len))) + "s -> %s"

  for item in sorted(args.__dict__.items()):
    if item[0] == "_args" and not include_raw_args:
      continue
    elif only_keys and not item[0] in only_keys:
      continue
    elif skip_keys and item[0] in skip_keys:
      continue

    logging.info(log_format, item[0], item[1])



#!/usr/bin/env python
# pylint: disable=C0103
import codecs
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
  user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

  def initialize_options(self):
    TestCommand.initialize_options(self)
    try:
      from multiprocessing import cpu_count

      self.pytest_args = ['-n', str(cpu_count()), '--boxed']
    except (ImportError, NotImplementedError):
      self.pytest_args = ['-n', '1', '--boxed']

  def finalize_options(self):
    TestCommand.finalize_options(self)
    self.test_args = []
    self.test_suite = True

  def run_tests(self):
    import pytest

    errno = pytest.main(self.pytest_args)
    sys.exit(errno)


requirements = codecs.open('./requirements.txt').read().splitlines()
long_description = codecs.open('./README.md').read()
version = codecs.open('./VERSION').read().strip()

test_requirements = [
  'pytest-cov',
  'pytest-mock',
  'pytest-xdist',
  'pytest',
]

setup(
  name='ltpylib',
  version=version,
  description='Common Python helper functions',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url='https://github.com/lancethomps/ltpylib',
  project_urls={
    'Bug Reports': 'https://github.com/lancethomps/ltpylib/issues',
    'Source': 'https://github.com/lancethomps/ltpylib',
  },
  author='Lance Thompson',
  license='MIT',
  keywords='utils',
  python_requires='>=3.6',
  packages=['ltpylib', 'ltpylibtests'],
  install_requires=requirements,
  classifiers=[],
  cmdclass={
    'test': PyTest,
  },
  tests_require=test_requirements,
  extras_require={
    "pg": ["psycopg2"],
  },
  entry_points={"console_scripts": ["get_python_available_functions = ltpylib.cli:get_python_available_functions",]},
)



#!/usr/bin/env python
# pylint: disable=C0111
from typing import Callable, List, Optional, TypeVar, Union

EMPTY_LIST: frozenset = frozenset([])
EMPTY_MAP: tuple = tuple(sorted({}.items()))
T = TypeVar("T")


def add_missing_to_list(main_list: list, others: list) -> list:
  if not main_list:
    return others
  elif not others:
    return main_list

  main_list.extend([val for val in others if val not in main_list])
  return main_list


def divide_chunks(values: list, chunk_size: int):
  for idx in range(0, len(values), chunk_size):
    yield values[idx:idx + chunk_size]


def flatten(list_of_lists: List[List]) -> List:
  from itertools import chain

  return list(chain.from_iterable(list_of_lists))


def flatten_list_of_possible_csv_strings(vals: List[str], sep: str = ",") -> List[str]:
  if not vals:
    return vals

  flattened = []
  for val in vals:
    flattened.extend(val.split(sep))

  return flattened


def modify_list_of_dicts(
  datas: List[dict],
  fn: Callable[[dict], Optional[dict]],
  in_place: bool = False,
) -> List[dict]:
  result: List[dict] = datas

  if not in_place:
    result = []

  for idx, data in enumerate(datas):
    if in_place:
      updated_data = data
    else:
      updated_data = data.copy()

    fn_result = fn(updated_data)
    if fn_result is not None:
      updated_data = fn_result

    if in_place:
      result[idx] = updated_data
    else:
      result.append(updated_data)

  return result


def remove_nulls(values: List[T]) -> List[T]:
  return [elem for elem in values if elem is not None]


def to_csv(values: Union[List, None], sep: str = ",") -> Union[str, None]:
  return sep.join(values) if values else None



#!/usr/bin/env python
import logging
import os.path
from getpass import getpass, getuser
from pathlib import Path
from typing import List, Union

from ltpylib import enums, inputs, procs

MAC_SOUND_DIRS = [
  "/System/Library/Sounds",
  "~/Library/Sounds",
]


class MacSoundsSystem(enums.EnumAutoName):
  BASSO = "Basso"
  BLOW = "Blow"
  BOTTLE = "Bottle"
  FROG = "Frog"
  FUNK = "Funk"
  GLASS = "Glass"
  HERO = "Hero"
  MORSE = "Morse"
  PING = "Ping"
  POP = "Pop"
  PURR = "Purr"
  SOSUMI = "Sosumi"
  SUBMARINE = "Submarine"
  TINK = "Tink"


MAC_SOUND_FAILURE = MacSoundsSystem.BASSO
MAC_SOUND_FINISHED = MacSoundsSystem.FUNK
MAC_SOUND_COPIED = MacSoundsSystem.SUBMARINE

COPIED_MSG = "Copied to clipboard"


def notify(message: str, title: str = "Terminal Notification", sound_name: Union[str, MacSoundsSystem] = MacSoundsSystem.PING, subtitle: str = ""):
  message = message.replace('"', '\\"').replace("\n", "\\n")
  title = title.replace('"', '\\"')
  sound_name = (sound_name.name if isinstance(sound_name, MacSoundsSystem) else sound_name).replace('"', '\\"')
  subtitle = subtitle.replace('"', '\\"')

  js_function = f"""
var app = Application.currentApplication()

app.includeStandardAdditions = true

app.displayNotification("{message}", {{
    withTitle: "{title}",
    subtitle: "{subtitle}",
    soundName: "{sound_name}"
}})
"""
  result = procs.run_with_regular_stdout([
    "/usr/bin/osascript",
    "-l",
    "JavaScript",
    "-e",
    js_function,
  ])
  if __name__ == "__main__":
    exit(result.returncode)


def pbcopy(
  val: str,
  log_copied: bool = False,
  log_sound: bool = False,
  sound: Union[MacSoundsSystem, str] = MAC_SOUND_COPIED,
  log_msg: str = COPIED_MSG,
):
  procs.run_with_regular_stdout(
    ["pbcopy"],
    input=val,
    check=True,
  )

  if log_copied:
    import logging

    if not logging.root.hasHandlers():
      print(log_msg)
    else:
      logging.info(log_msg)

  if log_sound:
    play_sound(sound)


def pbcopy_and_log(
  val: str,
  log_sound: bool = True,
  sound: Union[MacSoundsSystem, str] = MAC_SOUND_COPIED,
  log_msg: str = COPIED_MSG,
):
  pbcopy(
    val,
    log_copied=True,
    log_sound=log_sound,
    sound=sound,
    log_msg=log_msg,
  )


def pbpaste() -> str:
  return procs.run_and_parse_output_on_success(["pbpaste"])


def find_sound_file(sound: Union[MacSoundsSystem, str]):
  sound_name = (sound.name if isinstance(sound, MacSoundsSystem) else sound)
  for sounds_dir in MAC_SOUND_DIRS:
    for file_ext in ["", ".aiff"]:
      sound_file = Path(os.path.expanduser(sounds_dir)).joinpath(sound_name + file_ext)
      if sound_file.is_file():
        return sound_file

  if isinstance(sound, MacSoundsSystem):
    raise ValueError("sound file not found for: %s" % sound)

  sound_file = Path(os.path.expanduser(sound_name))
  if sound_file.is_file():
    return sound_file

  raise ValueError("sound file not found for: %s" % sound)


def play_sound(sound: Union[MacSoundsSystem, str]):
  procs.run_popen(["afplay", find_sound_file(sound).as_posix()])


def add_generic_password(label: str, pw: str, account: str = None) -> bool:
  if account is None:
    account = getuser()

  result = procs.run_with_regular_stdout([
    "security",
    "add-generic-password",
    "-a",
    account,
    "-s",
    label,
    "-w",
    pw,
  ])
  if result.returncode != 0:
    raise Exception("Could not add generic keychain password: label=%s account=%s" % (label, account))

  return True


def find_generic_password(
  label: str,
  ask_if_missing: bool = True,
  add_if_missing: bool = False,
  prompt_to_add_if_missing: bool = True,
) -> str:
  status, pw = procs.run_and_parse_output(
    [
      "security",
      "find-generic-password",
      "-ws",
      label,
    ]
  )
  if status != 0:
    if ask_if_missing:
      prompt = 'Please enter your password for %s.' % (label)
      if not add_if_missing:
        prompt += ' You can skip entering your password by running: security add-generic-password -a "$USER" -s "%s" -w' % (label)

      pw = getpass(prompt="%s\n> " % prompt)
      if add_if_missing and pw:
        if not prompt_to_add_if_missing or inputs.confirm('Add this password to your keychain in order to skip this prompt next time?', default="y"):
          logging.info('Adding generic password to keychain...')
          add_generic_password(label, pw, account=getuser())
    else:
      raise Exception("Could not find generic keychain password for label: %s" % label)

  return pw.strip()


def add_internet_password(host: str, pw: str, user: str = None) -> bool:
  if user is None:
    user = getuser()

  result = procs.run_with_regular_stdout([
    "security",
    "add-internet-password",
    "-t",
    "dflt",
    "-a",
    user,
    "-s",
    host,
    "-w",
    pw,
  ])
  if result.returncode != 0:
    raise Exception("Could not add internet keychain password: host=%s user=%s" % (host, user))

  return True


def find_internet_password(
  host: str,
  user: str = None,
  ask_if_missing: bool = True,
  add_if_missing: bool = False,
  prompt_to_add_if_missing: bool = True,
):
  if user is None:
    user = getuser()

  status, pw = procs.run_and_parse_output(
    [
      "/usr/bin/security",
      "find-internet-password",
      "-t",
      "dflt",
      "-a",
      user,
      "-ws",
      host,
    ]
  )
  if status != 0:
    if ask_if_missing:
      prompt = 'Please enter your password for %s.' % (host)
      if not add_if_missing:
        prompt += ' You can skip entering your password by running: security add-internet-password -a "$USER" -s "%s" -w' % (host)

      pw = getpass(prompt="%s\n> " % prompt)
      if add_if_missing and pw:
        if not prompt_to_add_if_missing or inputs.confirm('Add this password to your keychain in order to skip this prompt next time?', default="y"):
          logging.info('Adding internet password to keychain...')
          add_internet_password(host, pw, user=user)
    else:
      raise Exception("Could not find internet keychain password for label: %s" % host)

  return pw.strip()


def open_url_and_log(url: str, debug_mode: bool = False):
  return open_url(url, log_url=True, debug_mode=debug_mode)


def open_url(url: str, log_url: bool = False, debug_mode: bool = False):
  if log_url:
    logging.info(url)

  if debug_mode:
    return

  procs.run_with_regular_stdout(["open", url], check=True)


def trash(paths: Union[List[Path], List[str], Path, str], check: bool = True) -> bool:
  if not isinstance(paths, list):
    paths = [paths]

  paths = [p.as_posix() if isinstance(p, Path) else p for p in paths]

  return procs.run_with_regular_stdout(["trash", "-Fv"] + paths, check=check).returncode == 0


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
# pylint: disable=C0111

import re
import urllib.parse as urllib_parse
from typing import Callable, List, Union

StrConverter = Callable[[str], str]

SEARCH_SEP_CHARS: List[str] = [
  " ",
  "_",
  "-",
  ".",
]


def stripper(val: str) -> str:
  return val.strip()


def lower_case(val: str) -> str:
  return val.lower()


def alphanum_only(val: str) -> str:
  return re.sub(r'[^a-zA-Z0-9]', '', val)


def num_only(val: str) -> str:
  return re.sub(r'[^0-9.]', '', val)


def url_encode(val: str, safe: Union[bytes, str] = "", use_plus_for_space: bool = True) -> str:
  if not use_plus_for_space:
    return urllib_parse.quote(val, safe=safe)

  return urllib_parse.quote_plus(val, safe=safe)


def url_decode(val: str) -> str:
  return urllib_parse.unquote_plus(val)


def dict_to_url_params(params: dict, safe: str = "") -> str:
  return urllib_parse.urlencode(params, safe=safe)


def to_camel_case(val: str, sep: str = None) -> str:
  if not val:
    return val

  if sep is None:
    sep = find_sep_char(val)

  val = re.sub(r'[^a-zA-Z0-9 ]', ' ', val.lower().replace(sep, " ")).title()
  return (val[0:1].lower() + val[1:]).replace(" ", "")


def to_snake_case(val: str, sep: str = None) -> str:
  if not val:
    return val

  if sep is None:
    sep = find_sep_char(val)

  val = re.sub(r'[^a-zA-Z0-9 ]', ' ', val.lower().replace(sep, " "))
  return "_".join(val.lower().split())


def find_sep_char(val: str) -> str:
  if val.count(" ") > 0:
    return " "

  sep = ""
  count: int = -1
  for sep_char in SEARCH_SEP_CHARS:
    sep_char_count = val.count(sep_char)
    if sep_char_count > count:
      count = sep_char_count
      sep = sep_char

  return sep


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
from typing import List, Tuple, Union

import stashy
from requests import Response
from stashy import repos
from stashy.pullrequests import PullRequest, PullRequests
from stashy.repos import Repos

from ltpylib import colors, dates, requests_helper
from ltpylib.inputs import select_prompt_and_return_indexes
from ltpylib.stash_types import (
  Branch,
  Branches,
  Builds,
  PullRequestActivities,
  PullRequestMergeability,
  PullRequestMergeStatus,
  PullRequestParticipantStatus,
  PullRequestRole,
  PullRequestState,
  PullRequestStatus,
  PullRequestStatuses,
  Repository,
  SearchResults,
)
from ltpylib.strings import str_list_max_length

BRANCH_PR_META_KEY: str = "com.atlassian.bitbucket.server.bitbucket-ref-metadata:outgoing-pull-request-metadata"


class StashApi(object):

  def __init__(self, stash: stashy.Stash):
    self.stash: stashy.Stash = stash

  def ask_user_to_select_their_prs(
    self,
    role: PullRequestRole = None,
    state: PullRequestState = PullRequestState.OPEN,
    participant_status: List[PullRequestParticipantStatus] = None,
  ) -> Union[None, List[PullRequestStatus]]:
    my_prs = self.my_pull_requests(
      state=state,
      role=role,
      participant_status=participant_status,
    )

    if not my_prs.values:
      return None

    prs = sorted(my_prs.values, key=pr_sort)
    pr_links: List[str] = []

    branch_max_length = str_list_max_length([pr.fromRef.displayId for pr in prs])
    repo_max_length = str_list_max_length([pr.fromRef.repository.name for pr in prs])

    for pr in prs:
      pr_links.append(
        "  ".join([
          colors.green(dates.from_millis(pr.createdDate).isoformat()),
          colors.red(pr.fromRef.repository.name.ljust(repo_max_length)),
          colors.blue(pr.fromRef.displayId.ljust(branch_max_length)),
          pr.title,
        ])
      )

    selections = select_prompt_and_return_indexes(
      pr_links,
      header="Select PRs to open",
      multi=True,
      ansi=True,
    )

    return [prs[index] for index in selections]

  def builds_for_commit(self, commit: str, limit: int = 1000) -> Builds:
    api_url = '%s/rest/build-status/latest/commits/%s?limit=%s' % (
      self.stash._client._base_url,
      commit,
      str(limit),
    )
    kw = requests_helper.add_json_headers()
    result: Builds = Builds(requests_helper.parse_raw_response(self.stash._client._session.get(api_url, **kw)))

    return result

  def create_pull_request(
    self,
    project: str,
    repo: str,
    from_ref: str,
    to_ref: str,
    title: str,
    description: str,
    reviewers: List[str],
  ) -> PullRequestStatus:
    prs: PullRequests = self.stash.projects[project].repos[repo].pull_requests
    response: dict = prs.create(
      title=title,
      fromRef=from_ref,
      toRef=to_ref,
      description=description,
      reviewers=reviewers,
    )

    return PullRequestStatus(response)

  def delete_pull_request_source_branch(
    self,
    project: str,
    repo: str,
    pr_id: int,
  ) -> Response:
    api_url = '%s/rest/pull-request-cleanup/latest/projects/%s/repos/%s/pull-requests/%s' % (
      self.stash._client._base_url,
      project,
      repo,
      str(pr_id),
    )
    kw = requests_helper.add_json_headers()
    response = self.stash._client._session.post(
      api_url,
      json={
        "deleteSourceRef": True,
        "retargetDependents": True,
      },
      **kw,
    )
    return response

  def merge_pull_request(
    self,
    project: str,
    repo: str,
    pr_id: int,
    delete_source_branch: bool = True,
    commit_message: str = "",
    version: int = None,
  ) -> PullRequestMergeStatus:
    pr: PullRequest = self.stash.projects[project].repos[repo].pull_requests[str(pr_id)]
    if version is None:
      pr_info = self.pull_request(project, repo, pr_id, include_merge_info=False)
      version = pr_info.version

    result: PullRequestMergeStatus = PullRequestMergeStatus(
      requests_helper.parse_raw_response(pr._client.post(pr.url("/merge"), data=dict(message=commit_message, version=version)))
    )
    if delete_source_branch and result.state == "MERGED":
      delete_source_branch_response = self.delete_pull_request_source_branch(project, repo, pr_id)
      result.sourceBranchDeleted = delete_source_branch_response.status_code == 200 or delete_source_branch_response.status_code == 204

    return result

  def my_pull_requests(
    self,
    role: PullRequestRole = None,
    state: PullRequestState = PullRequestState.OPEN,
    participant_status: List[PullRequestParticipantStatus] = None,
    start: int = 0,
    limit: int = 100,
    with_attributes: bool = True,
    order: str = None,
  ) -> PullRequestStatuses:

    if isinstance(role, str):
      role = PullRequestRole.from_string(role, allow_unknown=True)

    if isinstance(state, str):
      state = PullRequestState.from_string(state, allow_unknown=True, unknown_as_none=True)

    if isinstance(participant_status, str):
      participant_status = [PullRequestParticipantStatus.from_string(val) for val in participant_status.split(",")] if participant_status and participant_status != "_" else None

    api_params = {
      "start": str(start),
      "limit": str(limit),
      "withAttributes": str(with_attributes).lower(),
    }

    if state:
      api_params["state"] = state.name

    if role:
      api_params["role"] = role.name

    if participant_status:
      api_params["participantStatus"] = ",".join([val.name for val in participant_status])

    if order:
      api_params["order"] = order

    api_url = '%s/rest/api/latest/dashboard/pull-requests?%s' % (
      self.stash._client._base_url,
      "&".join((k + "=" + v) for (k, v) in api_params.items()),
    )
    kw = requests_helper.add_json_headers()
    return PullRequestStatuses(values=requests_helper.parse_raw_response(self.stash._client._session.get(api_url, **kw)))

  def pull_request(
    self,
    project: str,
    repo: str,
    pr_id: int,
    include_merge_info: bool = True,
    include_builds: bool = False,
  ) -> PullRequestStatus:
    pr: PullRequest = self.stash.projects[project].repos[repo].pull_requests[str(pr_id)]
    result: PullRequestStatus = PullRequestStatus(pr.get())
    if include_merge_info and result.open:
      merge_info = self.pull_request_merge_info(project, repo, pr_id)
      result.mergeInfo = merge_info

    if include_builds and result.fromRef and result.fromRef.latestCommit:
      builds = self.builds_for_commit(result.fromRef.latestCommit)
      result.builds = builds

    return result

  def pull_request_activities(
    self,
    project: str,
    repo: str,
    pr_id: int,
  ) -> PullRequestActivities:
    pr: PullRequest = self.stash.projects[project].repos[repo].pull_requests[str(pr_id)]
    result: PullRequestActivities = PullRequestActivities(requests_helper.parse_raw_response(pr._client.get(pr.url() + "/activities")))

    return result

  def pull_request_merge_info(
    self,
    project: str,
    repo: str,
    pr_id: int,
  ) -> PullRequestMergeability:
    pr: PullRequest = self.stash.projects[project].repos[repo].pull_requests[str(pr_id)]
    return PullRequestMergeability(pr.merge_info())

  def pull_requests(
    self,
    project: str,
    repo: str,
    state: Union[PullRequestState, str] = PullRequestState.OPEN,
    author: str = None,
    order: str = None,
    target_branch: str = None,
    direction: str = 'INCOMING',
  ) -> List[PullRequestStatus]:

    if isinstance(state, str):
      state = PullRequestState.from_string(state, allow_unknown=False)

    prs: PullRequests = self.stash.projects[project].repos[repo].pull_requests
    return [PullRequestStatus(pr) for pr in prs.all(
      direction=direction,
      at=target_branch,
      state=(state.name if state else None),
      order=order,
      author=author,)]

  def repo(self, project: str, repo: str) -> Repository:
    return Repository(self.stash.projects[project].repos[repo].get())

  def repo_branches(self, project: str, repo: str, limit: int = None, details: bool = True) -> Branches:
    if isinstance(limit, str):
      limit = int(limit)
    elif limit is None:
      limit = 1000

    if isinstance(details, str):
      details = details.lower() == "true" or details.lower() == "yes"

    repo: repos.Repository = self.stash.projects[project].repos[repo]

    limit_above_max = limit > 1000 or limit <= -1
    if details and limit_above_max:
      branches = Branches(values={
        "limit": limit,
        "start": 0,
        "values": [],
      })
      branches.isLastPage = True

      for branch in repo.branches(details=details):
        branches.values.append(Branch(values=branch))
        if 0 <= limit <= len(branches.values):
          branches.isLastPage = False
          break

      branches.size = len(branches.values)
      return branches

    details_str = "true" if details else "false"
    if limit <= -1:
      limit = 999999

    return Branches(values=requests_helper.parse_raw_response(self.stash._client.get(repo.url('/branches?limit=%s&details=%s' % (limit, details_str)))))

  def repo_branches_with_pr_in_states(self, project: str, repo: str, states: List[PullRequestState], limit: int = None) -> Branches:
    if isinstance(limit, str):
      limit = int(limit)

    if isinstance(states, str):
      states = [PullRequestState.from_string(state) for state in states.split(",")]
    else:
      states = [PullRequestState.from_string(state) for state in states]

    branches = self.repo_branches(project, repo, limit=limit, details=True)

    merged_branches: Branches = Branches(values={
      "isLastPage": branches.isLastPage,
      "limit": branches.limit,
      "start": branches.start,
    })
    merged_branches.values = []

    for branch in branches.values:
      if branch.metadata is None or BRANCH_PR_META_KEY not in branch.metadata:
        continue

      branch_pr_meta = branch.metadata.get(BRANCH_PR_META_KEY)

      if "pullRequest" not in branch_pr_meta:
        has_prs_with_state = False
        for state in states:
          if state.name.lower() in branch_pr_meta and branch_pr_meta.get(state.name.lower()) > 0:
            has_prs_with_state = True
            break

        if has_prs_with_state:
          merged_branches.values.append(branch)

        continue

      pr_meta_raw: dict = branch_pr_meta.pop("pullRequest")
      pr_meta = PullRequestStatus(values=pr_meta_raw)
      branch_pr_meta["pullRequest"] = pr_meta
      if pr_meta.state in states:
        merged_branches.values.append(branch)

    merged_branches.size = len(merged_branches.values)
    return merged_branches

  def repo_merged_branches(self, project: str, repo: str, limit: int = None) -> Branches:
    return self.repo_branches_with_pr_in_states(project, repo, [PullRequestState.MERGED], limit=limit)

  def repos_for_project(self, project: str) -> List[Repository]:
    project_repos: Repos = self.stash.projects[project].repos
    return [Repository(repo) for repo in project_repos.list()]

  def search(self, query: str, limit: int = 25) -> SearchResults:
    api_json = {
      "entities": {
        "code": {}
      },
      "limits": {
        "primary": limit
      },
      "query": query,
    }
    kw = requests_helper.add_json_headers()
    return SearchResults(requests_helper.parse_raw_response(self.stash._client._session.post(
      '%s/rest/search/latest/search' % self.stash._client._base_url,
      json=api_json,
      **kw,
    )))


def create_stash_api(url: str, creds: Tuple[str, str] = None, token: str = None) -> StashApi:
  if token:
    return StashApi(stashy.client.Stash(url, token=token))

  return StashApi(stashy.client.Stash(url, username=creds[0], password=creds[1]))


def pr_sort(pr: PullRequestStatus) -> str:
  return "%s %s" % (
    pr.fromRef.repository.name,
    pr.fromRef.displayId.lower(),
  )



#!/usr/bin/env python
import atexit
from typing import List, Optional, Type, TypeVar

import datadog_api_client
import requests
from datadog_api_client.model_utils import deserialize_model, OpenApiModel
from datadog_api_client.v1.api.dashboards_api import DashboardsApi
from datadog_api_client.v1.api.monitors_api import MonitorsApi
from datadog_api_client.v1.model.monitor import Monitor
from datadog_api_client.v1.model.monitor_update_request import MonitorUpdateRequest

from ltpylib import requests_helper

DD_API_BASE = "https://api.datadoghq.com/api/v1"
OpenApiModelType = TypeVar("OpenApiModelType", bound=OpenApiModel)


class DatadogApi(object):

  def __init__(
    self,
    dd_api_key: str,
    dd_application_key: str = None,
    base_url: str = DD_API_BASE,
  ):
    self._dd_api_key: str = dd_api_key
    self._dd_application_key: str = dd_application_key

    self.dd_api_client: datadog_api_client.ApiClient = datadog_api_client.ApiClient(
      datadog_api_client.Configuration(api_key={
        "apiKeyAuth": dd_api_key,
        "appKeyAuth": dd_application_key,
      })
    )
    atexit.register(self.dd_api_client.close)
    self.dashboards_api: DashboardsApi = DashboardsApi(self.dd_api_client)
    self.monitors_api: MonitorsApi = MonitorsApi(self.dd_api_client)

    self.base_url: str = base_url.removesuffix("/")
    self.session: requests.Session = requests.Session()
    self.session.headers.update({
      "Accept": "application/json",
      "Content-Type": "application/json",
      "DD-API-KEY": dd_api_key,
    })
    if dd_application_key:
      self.session.headers.update({"DD-APPLICATION-KEY": dd_application_key})

  def create_api_model(self, data: dict, model_type: Type[OpenApiModelType], check_type: bool = True) -> OpenApiModelType:
    return deserialize_model(data, model_type, [], check_type, self.dd_api_client.configuration, True)

  def url(self, path: str) -> str:
    if not path.startswith("/"):
      path = "/" + path

    return self.base_url + path

  def get_dashboard(self, dashboard_id: str) -> dict:
    return requests_helper.maybe_throw(self.session.get(self.url(f"/dashboard/{dashboard_id}"))).json()

  def get_dashboards(self) -> List[dict]:
    return requests_helper.maybe_throw(self.session.get(self.url("/dashboard"))).json()["dashboards"]

  def update_dashboard(self, dashboard_id: str, dashboard: dict) -> dict:
    return requests_helper.maybe_throw(self.session.put(self.url(f"/dashboard/{dashboard_id}"), json=dashboard)).json()

  def get_monitors(self, monitor_tags: Optional[str] = None) -> List[dict]:
    params = {}
    if monitor_tags:
      params["monitor_tags"] = monitor_tags

    return requests_helper.maybe_throw(self.session.get(self.url("/monitor"), params=params)).json()

  def update_monitor(self, monitor_id: int, monitor: dict) -> Monitor:
    return self.monitors_api.update_monitor(monitor_id, self.create_api_model(monitor, MonitorUpdateRequest))



#!/usr/bin/env python
import base64
import json
from typing import Dict, Optional, Tuple

from boto3.session import Session
from botocore.exceptions import ClientError

AWS_SESSIONS_BY_REGION_PROFILE: Dict[Tuple[Optional[str], Optional[str]], Session] = {}


def create_aws_session(
  region_name: str = None,
  profile_name: str = None,
) -> Session:
  return Session(region_name=region_name, profile_name=profile_name)


def get_or_create_aws_session(
  region_name: str = None,
  profile_name: str = None,
) -> Session:
  key = (region_name, profile_name)
  if key in AWS_SESSIONS_BY_REGION_PROFILE:
    return AWS_SESSIONS_BY_REGION_PROFILE[key]

  session = create_aws_session(region_name=region_name, profile_name=profile_name)
  AWS_SESSIONS_BY_REGION_PROFILE[key] = session
  return session


def set_default_aws_session(region_name: str = None, profile_name: str = None) -> Session:
  session = get_or_create_aws_session(region_name=region_name, profile_name=profile_name)
  AWS_SESSIONS_BY_REGION_PROFILE[(None, None)] = session
  return session


def get_secret_json(secret_name: str, session: Session = None) -> dict:
  if session is None:
    session = get_or_create_aws_session()

  client = session.client(service_name="secretsmanager")

  try:
    get_secret_value_response = client.get_secret_value(SecretId=secret_name)
  except ClientError as e:
    if e.response["Error"]["Code"] == "DecryptionFailureException":
      # Secrets Manager can"t decrypt the protected secret text using the provided KMS key.
      # Deal with the exception here, and/or rethrow at your discretion.
      raise e
    elif e.response["Error"]["Code"] == "InternalServiceErrorException":
      # An error occurred on the server side.
      # Deal with the exception here, and/or rethrow at your discretion.
      raise e
    elif e.response["Error"]["Code"] == "InvalidParameterException":
      # You provided an invalid value for a parameter.
      # Deal with the exception here, and/or rethrow at your discretion.
      raise e
    elif e.response["Error"]["Code"] == "InvalidRequestException":
      # You provided a parameter value that is not valid for the current state of the resource.
      # Deal with the exception here, and/or rethrow at your discretion.
      raise e
    elif e.response["Error"]["Code"] == "ResourceNotFoundException":
      # We can"t find the resource that you asked for.
      # Deal with the exception here, and/or rethrow at your discretion.
      raise e
  else:
    # Decrypts secret using the associated KMS key.
    # Depending on whether the secret is a string or binary, one of these fields will be populated.
    if "SecretString" in get_secret_value_response:
      secret = get_secret_value_response["SecretString"]
    else:
      secret = base64.b64decode(get_secret_value_response["SecretBinary"])

    return json.loads(secret)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
import argparse

from ltpylib import opts


def get_python_available_functions():
  import importlib
  from ltpylib import reflections

  arg_parser = opts.create_default_with_positionals_arg_parser(positionals_key="module")
  arg_parser.add_argument("--include-signature", "-is", action=argparse.BooleanOptionalAction, default=False)
  arg_parser.add_argument("--attr", "--class-name", "-a", "-c", help="Attribute to get functions from, such as a class name.")
  arg_parser.add_argument("--sep-char", "-s", default=" ")

  args = opts.parse_args_with_positionals_and_init_others(arg_parser, log_use_stderr=True)

  imported = importlib.import_module(args.module[0])
  if args.attr:
    imported = getattr(imported, args.attr)

  funcs = reflections.get_functions_of_class(imported, include_signature=args.include_signature)
  print(args.sep_char.join(funcs))



#!/usr/bin/env python
import os
from pathlib import Path
from typing import Any, Generic, Sequence, TypeVar, Union

import looker_sdk
from looker_sdk.rtl.serialize import TModelOrSequence
from looker_sdk.sdk.api40.methods import Looker40SDK
from looker_sdk.sdk.api40.models import Dashboard, Look, LookWithQuery

from ltpylib import files
from ltpylib.output import prettify_json, prettify_json_compact

LOOKER_CONFIG_FILE_DEFAULT = Path.home().joinpath(".config/looker.ini")
T = TypeVar("T")


class LookerApi(object):

  def __init__(
    self,
    config_file: Path = LOOKER_CONFIG_FILE_DEFAULT,
    user_id: str = os.getenv("LOOKER_USER_ID"),
  ):
    self.config_file: Path = config_file
    self.user_id: str = user_id

    self.client: Looker40SDK = looker_sdk.init40(config_file=self.config_file.as_posix())

  @staticmethod
  def deserialize(data: Union[str, Path, Any], structure: Generic[T]) -> T:
    if isinstance(data, Path):
      data = files.read_file(data)
    elif not isinstance(data, str):
      data = prettify_json_compact(data)

    return looker_sdk.rtl.serialize.deserialize40(data=data, structure=structure)

  @staticmethod
  def remove_dashboard_non_config_fields(config: Dashboard) -> Dashboard:
    config.last_accessed_at = None
    config.last_viewed_at = None
    config.view_count = None
    return config

  @staticmethod
  def remove_config_non_config_fields(config: Union[Dashboard, Look, LookWithQuery]) -> Union[Dashboard, Look, LookWithQuery]:
    config.last_accessed_at = None
    config.last_viewed_at = None
    config.view_count = None
    return config

  @staticmethod
  def to_json(api_model: TModelOrSequence) -> str:
    return prettify_json(looker_sdk.rtl.serialize.converter40.unstructure(api_model))

  @property
  def user_id(self) -> str:
    if not self._user_id:
      self._user_id = self.client.me().id

    return self._user_id

  @user_id.setter
  def user_id(self, user_id: str):
    self._user_id = user_id

  def my_dashboards(self, limit: int = 100) -> Sequence[Dashboard]:
    return self.client.search_dashboards(user_id=self.user_id, limit=limit)

  def my_looks(self, limit: int = 100) -> Sequence[Look]:
    return self.client.search_looks(user_id=self.user_id, limit=limit)
